import logging
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable, Protocol

from jira import JIRAError
from tqdm import tqdm

from jf_ingest.data_manifests.jira.adapters.manifest_adapter import ManifestAdapter
from jf_ingest.data_manifests.jira.manifest_base import (
    IJiraDataManifest,
    IJiraProjectManifest,
)
from jf_ingest.data_manifests.manifest_base import ManifestSource

logger = logging.getLogger(__name__)


class UnsupportedJiraProvider(Exception):
    pass


class ProjectManifestGenerationException(Exception):
    pass


# This function was moved from Jellyfish core along with base models for Jira manifests.
# It is used to generate a Jira manifest from a manifest adapter.
# The manifest adapter is a class that provides the necessary data to generate the manifest.
# create_jira_data_manifest and create_jira_project_manifest are callables that create the actual manifest objects.
#  This was done to retain the behavior of the original data/project manifests while avoiding circular imports.
# reset_current_tenant is a callable that resets the current tenant in the Jellyfish ORM when spawning threads.
#  This is passed in because jf_ingest has no concept of Django ORM and to avoid a circular import.
def create_manifest(
    manifest_adapter: ManifestAdapter,
    classifications_name_lookup,
    reset_current_tenant: Callable,
    create_jira_data_manifest: Callable,
    create_jira_project_manifest: Callable,
) -> type[IJiraDataManifest]:
    config = manifest_adapter.config
    project_data_dicts = manifest_adapter.get_project_data_dicts()

    jira_manifest = None
    # Total threads includes all potential project manifests
    # plus one 'global' jira manifest thread that will get
    # all totals (sprints, boards, issues, etc)
    total_threads = len(project_data_dicts) + 1
    with tqdm(total=total_threads) as pbar:
        with ThreadPoolExecutor(max_workers=config.issue_download_concurrent_threads) as executor:
            pbar.set_description('Processing Jira Projects')

            def _update_pbar_from_future(future):
                pbar.update(1)

            # Create future for global Jira Manifest. This will be our main JiraManifest,
            # and we will add Project Manifests to it
            generate_base_manifest_future: Any = executor.submit(
                process_global_jira_data,
                manifest_adapter,
                create_jira_data_manifest,
                reset_current_tenant,
            )
            # Add callback
            generate_base_manifest_future.add_done_callback(_update_pbar_from_future)

            # generate futures for project manifests
            project_future_to_project_key: dict = {
                executor.submit(
                    generate_project_manifest,
                    manifest_adapter,
                    project_data_dict,
                    classifications_name_lookup,
                    create_jira_project_manifest,
                    reset_current_tenant,
                ): project_data_dict['key']
                for project_data_dict in project_data_dicts
            }

            # Add callbacks for project manifest futures
            for f in project_future_to_project_key.keys():
                f.add_done_callback(_update_pbar_from_future)

            # Process future results and handle exceptions
            project_keys_to_errors = {}
            project_manifests: list[IJiraDataManifest] = []
            for future in project_future_to_project_key.keys():
                # Track exceptions and log them in a dictionary
                # so we can map project keys to detected exceptions
                if future.exception():
                    project_keys_to_errors[project_future_to_project_key[future]] = (
                        future.exception()
                    )
                else:
                    project_manifests.append(future.result())

            pbar.set_description('Processing Jira Global Values')

            # Load base Jira Manifest object from globals generator
            jira_manifest = generate_base_manifest_future.result()

            # For sorting/typing, we can't allow None values as a key.
            def _get_project_manifests_project_key(project_manifest: IJiraProjectManifest) -> str:
                project_key: str = project_manifest.project_key or ''
                return project_key

            # Append additional manifests (only project manifests, counts related to excluded project manifest, etc)
            jira_manifest.project_manifests = sorted(
                project_manifests, key=_get_project_manifests_project_key
            )
            jira_manifest.encountered_errors_for_projects = project_keys_to_errors

            pbar.set_description('Done!')

    logger.info('Done processing Jira Manifest!')
    return_value: type[IJiraDataManifest] = jira_manifest
    return return_value


class CreateJiraProjectManifestProtocol(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> type[IJiraProjectManifest]: ...


class CreateJiraDataManifestProtocol(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> type[IJiraDataManifest]: ...


def generate_project_manifest(
    manifest_adapter: ManifestAdapter,
    project_data_dict: dict,
    classifications_name_lookup: dict,
    create_jira_project_manifest: CreateJiraProjectManifestProtocol,
    reset_current_tenant: Callable,
) -> type[IJiraProjectManifest]:
    # Need to reset tenancy when spinning up threads
    reset_current_tenant()

    project_key = project_data_dict['key']
    project_id = int(project_data_dict['id'])
    try:
        # FIRST, DO A BASIC TEST TO SEE IF WE HAVE THE PROPER PERMISSIONS
        # TO SEE ANY DATA IN THIS PROJECT
        if not manifest_adapter.test_basic_auth_for_project(project_id=project_id):
            raise ProjectManifestGenerationException(
                f'Authentication Exception Encountered for Project {project_key}!'
            )

        manifest_source = manifest_adapter.manifest_source
        excluded: bool = project_key in manifest_adapter.excluded_project_keys

        # Set the issues count to 0 if we are in a remote manifest
        # and this project is excluded
        issues_count = (
            0
            if manifest_source == ManifestSource.remote and excluded
            else manifest_adapter.get_issues_count_for_project(project_id=project_id)
        )
        version_count = manifest_adapter.get_project_versions_count_for_project(
            project_id=project_id
        )
        last_issue_updated_date = manifest_adapter.get_last_updated_for_project(
            project_id=project_id
        )
        project_classification = manifest_adapter.project_keys_to_classification_type.get(
            project_key
        )

        return create_jira_project_manifest(
            company=manifest_adapter.company_slug,
            data_source=manifest_source,
            project_id=str(project_id),
            project_key=project_key,
            issues_count=issues_count,
            version_count=version_count,
            excluded=excluded,
            pull_from=manifest_adapter.config.pull_from,
            last_issue_updated_date=last_issue_updated_date,
            classification=project_classification,
            classification_str=classifications_name_lookup.get(project_classification),
        )
    except JIRAError as e:
        logger.debug(f'{traceback.format_exc()}')
        raise ProjectManifestGenerationException(e.text)
    except Exception as e:
        logger.debug(f'{traceback.format_exc()}')
        raise ProjectManifestGenerationException(str(e))


def process_global_jira_data(
    manifest_adapter: ManifestAdapter,
    create_jira_data_manifest: CreateJiraDataManifestProtocol,
    reset_current_tenant: Callable,
) -> type[IJiraDataManifest]:
    # Need to reset tenancy when spinning up threads
    reset_current_tenant()

    total_users_count = manifest_adapter.get_users_count()
    total_fields_count = manifest_adapter.get_fields_count()
    total_resolutions_count = manifest_adapter.get_resolutions_count()
    total_issue_types_count = manifest_adapter.get_issue_types_count()
    total_issue_link_types_count = manifest_adapter.get_issue_link_types_count()
    total_priorities_count = manifest_adapter.get_priorities_count()
    total_boards_count = manifest_adapter.get_boards_count()
    project_data_dicts = manifest_adapter.get_project_data_dicts()
    project_versions_count = manifest_adapter.get_project_versions_count()
    issues_count = manifest_adapter.get_issues_count()

    return create_jira_data_manifest(
        company=manifest_adapter.company_slug,
        data_source=manifest_adapter.manifest_source,
        pull_from=manifest_adapter.config.pull_from,
        users_count=total_users_count,
        fields_count=total_fields_count,
        resolutions_count=total_resolutions_count,
        issue_types_count=total_issue_types_count,
        issue_link_types_count=total_issue_link_types_count,
        priorities_count=total_priorities_count,
        projects_count=len(project_data_dicts),
        boards_count=total_boards_count,
        project_versions_count=project_versions_count,
        issues_count=issues_count,
        # The following fields must be filled out after processing
        # all ProjectManifests
        project_manifests=[],
        encountered_errors_for_projects={},
    )



import inspect
import logging
import time
import traceback
from contextlib import contextmanager
from functools import wraps
from typing import Any, Optional

logger = logging.getLogger(__name__)

AGENT_LOG_TAG = 'agent_log_file_only'
LOG_FILE_NAME = "jf_agent.log"


def _get_logger_from_parent() -> logging.Logger:
    """
    Retrieves a logger instance, inheriting the parent's logger if available.
    """
    name = None
    current_frame = inspect.currentframe()

    # Grab the frame two levels up to get the calling module's name
    # I don't think we need excessive null checking here like this but I am paranoid
    if current_frame and current_frame.f_back and current_frame.f_back.f_back:
        frame = current_frame.f_back.f_back
        name = frame.f_globals.get('__name__', 'root')

    if name:
        return logging.getLogger(name)
    else:
        return logger


def log_entry_exit(*args, **kwargs):
    logger = _get_logger_from_parent()

    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            func_name = func.__name__
            start_time = time.time()
            logger.debug(f"{func_name}: Starting")

            ret = func(*args, **kwargs)

            logger.debug(f"{func_name}: Ending")
            end_time = time.time()
            elapsed_time = end_time - start_time
            logger.debug(f"{func_name}: Execution took {elapsed_time:.4f} seconds")
            return ret

        return wrapper

    return actual_decorator


@contextmanager
def log_loop_iters(loop_desc, this_iternum, log_every_n_iters, log_entry=True, log_exit=True):
    logger = _get_logger_from_parent()
    if (this_iternum - 1) % log_every_n_iters == 0:
        if log_entry:
            logger.debug(f'Loop "{loop_desc}", iter {this_iternum}: Starting')
        yield
        if log_exit:
            logger.debug(f'Loop "{loop_desc}", iter {this_iternum}: Ending')
    else:
        yield


# Mapping of error/warning codes to templated error messages to be called by
# log_standard_error(). This allows for Jellyfish to better categorize errors/warnings.
ERROR_MESSAGES = {
    0000: "An unknown error has occurred. Error message: {}",
    3000: "Failed to upload file {} to S3 bucket",
    3001: "Connection to bucket was disconnected while uploading: {} with the following error: {}. Retrying...",
    3010: "Rate limiter: thought we were operating within our limit (made {}/{} calls for {}), but got HTTP 429 anyway!",
    3020: "Next available time to make call is after the timeout of {} seconds. Giving up.",
    3030: "ERROR: Could not parse response with status code {}. Contact an administrator for help.",
    3011: "Error normalizing PR {} from repo {}. Skipping...",
    3021: "Error getting PRs for repo {}. Skipping...",
    3031: "Unable to parse the diff For PR {} in repo {}; proceeding as though no files were changed.",
    3041: "For PR {} in repo {}, caught HTTPError (HTTP 401) when attempting to retrieve changes; "
    "proceeding as though no files were changed",
    3051: "For PR {} in repo {}, caught UnicodeDecodeError when attempting to decode changes; proceeding as though no files were changed",
    3061: "Failed to download {} data:\n{}",
    3071: "Received retryable requests error ({}), retry {} / {}.  Trying again in {}...",
    3081: "Got unexpected HTTP 403 for repo {}.  Skipping...",
    3091: "Github rate limit exceeded.  Trying again in {}...",
    3101: "Request to {} has failed {} times -- giving up!",
    3121: 'Got HTTP {} when fetching commit {} for "{}", this likely means you are trying to fetch an invalid re',
    3131: "Got {} {} when {} ({})",
    3141: "Got {} {} when {}",
    3151: "Getting HTTP 429s for over an hour; giving up!",
    3002: "Failed to download jira data:\n{}",
    3012: "Caught KeyError from search_issues(), reducing batch size to {}",
    3022: "Caught KeyError from search_issues(), batch size is already 0, bailing out",
    3032: "Exception encountered in thread {}\n{}",
    3042: "[Thread {}] Jira issue downloader FAILED",
    3052: "JIRAError ({}), reducing batch size to {}",
    3062: "Apparently unable to fetch issue based on search_params {}",
    3082: "OJ-9084: Changelog history item with no 'fieldId' or 'field' key: {}",
    3092: (
        "OJ-22511: server side 500, batch size reduced to 0, "
        "last error was: {} with jql: {}, start_at: {}, and batch_size: {}. Skipping one issue ahead..."
    ),
    2101: "Failed to connect to {}:\n{}",
    2102: "Unable to access project {}, may be a Jira misconfiguration. Skipping...",
    2112: "Failed to connect to Jira for project ID: {}",
    2122: "you do not have the required 'development field' permissions in jira required to scan for missing repos",
    2132: (
        "Missing recommended jira_fields! For the best possible experience, "
        "please add the following to `include_fields` in the "
        "configuration file: {}"
    ),
    2142: (
        "Excluding recommended jira_fields! For the best possible experience, "
        "please remove the following from `exclude_fields` in the "
        "configuration file: {}",
    ),
    2201: (
        "\nERROR: Failed to download ({}) repo(s) from the group {}. "
        "Please check that the appropriate permissions are set for the following repos... ({})"
    ),
    2202: "You do not have the required permissions in jira required to fetch boards for the project {}",
    2203: "ERROR: Failed downloading sprints for Jira board: {} with s_start_at={}.\nReceived 400 response:\n{}",
}


def generate_standard_error_msg(error_code, msg_args=[]):
    return f"[{error_code}] {ERROR_MESSAGES.get(error_code).format(*msg_args)}"


def log_standard_error(level, error_code, msg_args=[], extra=None, exc_info=False):
    """
    For a failure that should be sent to the logger with an error_code, and also written
    to stdout (for user visibility)
    """
    logger = _get_logger_from_parent()
    assert level >= logging.WARNING
    msg = generate_standard_error_msg(error_code=error_code, msg_args=msg_args)
    logger.log(level=level, msg=msg, extra=extra)
    if exc_info:
        logger.log(level=level, msg=traceback.format_exc())


def send_to_agent_log_file(
    msg: str,
    level: int = logging.INFO,
    extra: Optional[dict] = None,
    exc_info: Optional[bool] = False,
):
    """Helper function for logging things to debug or INFO. Agent relies heavily on logging things to
    debug to send extra info to the agent.log file. For Direct Connect we'd like to get all the debug
    info as well, but due to prefect weirdness we have to do this work around.

    All logs submitted via this wrapper will be tagged as 'agent_log_file_only': True in the log records
    'extra' field

    Args:
        msg (str): Message to log
        level (int): The level of the log to log. Defaults to logging.INFO
        extra (dict): A dictionary of extra things to tag this log with. Defaults to {}
        exc_info (bool): Whether to log exception info. Defaults to False
    """
    log_extras: dict[str, Any] = {}
    logger = _get_logger_from_parent()

    if extra:
        log_extras.update(extra)

    log_extras[AGENT_LOG_TAG] = True
    logger.log(level=level, msg=msg, extra=log_extras, exc_info=exc_info)



from dataclasses import dataclass, field
from datetime import datetime
from enum import IntEnum
from functools import wraps
from typing import Any, Dict, List, Optional

from jf_ingest.jf_git.utils import (
    branch_redactor,
    organization_redactor,
    repo_redactor,
    sanitize_text,
)


@dataclass
class StandardizedObject:
    def redact_names_and_urls(self):
        '''This method is used to redact names and urls from the object,
        it should be implemented by the child classes'''
        pass

    def strip_text_content(self):
        '''This method is used to strip text content from the object,
        it should be implemented by the child classes'''
        pass


@dataclass
class StandardizedUser(StandardizedObject):
    id: str
    name: Optional[str]
    login: str
    email: Optional[str] = None
    url: Optional[str] = None
    account_id: Optional[str] = None


@dataclass
class StandardizedTeam(StandardizedObject):
    id: str
    slug: str
    name: str
    description: Optional[str]
    members: list[StandardizedUser]


@dataclass
class StandardizedBranch(StandardizedObject):
    name: Optional[str]
    sha: Optional[str]
    repo_id: str
    is_default: bool

    def redact_names_and_urls(self):
        self.name = branch_redactor.redact_name(self.name)


@dataclass
class StandardizedOrganization(StandardizedObject):
    id: str
    name: Optional[str]
    login: str
    url: Optional[str]

    def redact_names_and_urls(self):
        self.name = organization_redactor.redact_name(self.name)
        self.url = None


@dataclass
class StandardizedShortRepository(StandardizedObject):
    id: str
    name: str
    url: str

    def redact_names_and_urls(self):
        self.name = repo_redactor.redact_name(self.name)
        self.url = None


@dataclass
class StandardizedRepository(StandardizedObject):
    id: str
    name: str
    full_name: str
    url: str
    is_fork: bool
    default_branch_name: Optional[str]
    default_branch_sha: Optional[str]
    organization: StandardizedOrganization
    branches: list = field(default_factory=list)
    commits_backpopulated_to: Optional[datetime] = None
    prs_backpopulated_to: Optional[datetime] = None
    full_path: Optional[str] = (
        None  # This is only used by the Gitlab adapter and is ignored during git import
    )

    def redact_names_and_urls(self):
        self.name = repo_redactor.redact_name(self.name)
        self.full_name = repo_redactor.redact_name(self.full_name)
        self.url = None

        if self.default_branch_name:
            self.default_branch_name = branch_redactor.redact_name(self.default_branch_name)

        self.organization.redact_names_and_urls()

    def short(self):
        # return the short form of Standardized Repository
        return StandardizedShortRepository(id=self.id, name=self.name, url=self.url)


@dataclass
class StandardizedCommit(StandardizedObject):
    hash: str
    url: str
    message: str
    commit_date: datetime
    author_date: datetime
    author: Optional[StandardizedUser]
    repo: StandardizedShortRepository
    is_merge: bool
    branch_name: Optional[str] = None

    def redact_names_and_urls(self):
        self.branch_name = branch_redactor.redact_name(self.branch_name)
        self.url = None
        self.repo.redact_names_and_urls()

    def strip_text_content(self):
        self.message = sanitize_text(self.message)


@dataclass
class StandardizedPullRequestComment(StandardizedObject):
    user: Optional[StandardizedUser]
    body: str
    created_at: datetime
    system_generated: Optional[bool] = None

    def strip_text_content(self):
        self.body = sanitize_text(self.body)


# TODO: This exists in the Jellyfish source code
# and has been copied over. The source code should be
# replaced with this, so we have one "source of truth".
# NOTE: These enum string values are based off of Github,
# we need to normalize all other providers to this
class PullRequestReviewState(IntEnum):
    UNKNOWN = 0
    PENDING = 1
    APPROVED = 2
    COMMENTED = 3
    CHANGES_REQUESTED = 4
    DISMISSED = 5


@dataclass
class StandardizedPullRequestReview(StandardizedObject):
    user: Optional[StandardizedUser]
    foreign_id: str
    review_state: str


@dataclass
class StandardizedLabel(StandardizedObject):
    id: int
    name: str
    default: bool
    description: str


@dataclass
class StandardizedFileData(StandardizedObject):
    status: str
    changes: int
    additions: int
    deletions: int


@dataclass
class StandardizedPullRequest(StandardizedObject):
    id: Any
    additions: int
    deletions: int
    changed_files: int
    is_closed: bool
    is_merged: bool
    created_at: datetime
    updated_at: datetime
    merge_date: Optional[datetime]
    closed_date: Optional[datetime]
    title: str
    body: str
    url: str
    base_branch: str
    head_branch: str
    author: StandardizedUser
    merged_by: Optional[StandardizedUser]
    commits: List[StandardizedCommit]
    merge_commit: Optional[StandardizedCommit]
    comments: List[StandardizedPullRequestComment]
    approvals: List[StandardizedPullRequestReview]
    base_repo: StandardizedShortRepository
    head_repo: StandardizedShortRepository
    labels: List[StandardizedLabel]
    files: Dict[str, StandardizedFileData]

    def redact_names_and_urls(self):
        self.base_branch = branch_redactor.redact_name(self.base_branch)
        self.head_branch = branch_redactor.redact_name(self.head_branch)
        self.url = None

        self.base_repo.redact_names_and_urls()
        self.head_repo.redact_names_and_urls()

        if self.merge_commit:
            self.merge_commit.redact_names_and_urls()

        for commit in self.commits:
            commit.redact_names_and_urls()

    def strip_text_content(self):
        self.title = sanitize_text(self.title)
        self.body = sanitize_text(self.body)

        for comment in self.comments:
            comment.strip_text_content()

        for commit in self.commits:
            commit.strip_text_content()

        if self.merge_commit:
            self.merge_commit.strip_text_content()


@dataclass
class StandardizedPullRequestMetadata(StandardizedObject):
    id: Optional[Any]
    updated_at: datetime
    # API Index place holder, needed if the adapter DOES NOT support PullRequest API Time filtering
    # (git_provider_pr_endpoint_supports_date_filtering)
    api_index: Optional[Any] = None


@dataclass
class StandardizedJFAPIPullRequest(StandardizedObject):
    pr_number: int
    repo_name: str
    org_login: str


@dataclass
class StandardizedPullRequestAuthor(StandardizedObject):
    # PR fields
    pr_number: int
    repo_name: str
    org_login: str

    # User fields
    # Not using StandardizedUser here to allow for less strict type
    # definitions, specifically when our search could result in users
    # that have been deleted from their GHC instance
    user_id: Optional[str] = None
    user_login: Optional[str] = None
    user_name: Optional[str] = None
    user_email: Optional[str] = None


@dataclass
class StandardizedPullRequestReviewAuthor(StandardizedObject):
    # PR fields
    pr_number: int
    repo_name: str
    org_login: str

    # Review ID is the GraphQL DB ID, NOT the Node ID
    review_id: str

    # User fields
    user_id: Optional[str] = None
    user_login: Optional[str] = None
    user_name: Optional[str] = None
    user_email: Optional[str] = None



import fnmatch
import logging
import traceback
from abc import ABC, abstractmethod
from dataclasses import asdict
from datetime import datetime, timedelta, timezone
from enum import Enum
from itertools import chain
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
)

from requests.exceptions import HTTPError

from jf_ingest import diagnostics, logging_helper
from jf_ingest.config import (
    GitConfig,
    GitProvider,
    GitProviderInJellyfishRepo,
    IngestionConfig,
)
from jf_ingest.constants import Constants
from jf_ingest.events.models import GitIngestEvent, GitIngestionEventNames, IngestType
from jf_ingest.file_operations import IngestIOHelper, SubDirectory
from jf_ingest.jf_git.exceptions import (
    GitAuthorizationException,
    GitProviderUnavailable,
)
from jf_ingest.jf_git.jf_api import get_jf_github_null_user_pr_data
from jf_ingest.jf_git.standardized_models import (
    StandardizedBranch,
    StandardizedCommit,
    StandardizedJFAPIPullRequest,
    StandardizedObject,
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedPullRequestAuthor,
    StandardizedPullRequestMetadata,
    StandardizedPullRequestReviewAuthor,
    StandardizedRepository,
    StandardizedTeam,
    StandardizedUser,
)
from jf_ingest.telemetry import add_telemetry_fields, jelly_trace, record_span
from jf_ingest.utils import (
    batch_iterable,
    batch_iterable_by_bytes_size,
    get_jellyfish_company_slug,
    init_jf_ingest_run,
    tqdm_to_logger,
)

logger = logging.getLogger(__name__)

'''

    Constants

'''
# NOTE: ONLY GITHUB IS CURRENTLY SUPPORTED!!!!
BBS_PROVIDER = 'bitbucket_server'
BBC_PROVIDER = 'bitbucket_cloud'
GH_PROVIDER = 'github'
GL_PROVIDER = 'gitlab'
PROVIDERS = [GL_PROVIDER, GH_PROVIDER, BBS_PROVIDER, BBC_PROVIDER]


class BackpopulationWindow(NamedTuple):
    backpopulation_window_start: datetime
    backpopulation_window_end: datetime


class JFIngestGitProviderException(Exception):
    pass


class GitObject(Enum):
    GitOrganizations = "git_data_organizations"
    GitUsers = "git_data_users"
    GitTeams = "git_data_teams"
    GitRepositories = "git_data_repos"
    GitBranches = "git_data_branches"
    GitCommits = "git_data_commits"
    GitPullRequests = "git_data_prs"
    GitPullRequestAuthors = "git_data_pr_authors"
    GitPullRequestReviewAuthors = "git_data_pr_review_authors"


def _generate_git_ingest_event(event_name: str, git_config: GitConfig) -> GitIngestEvent:
    return GitIngestEvent(
        company_slug=get_jellyfish_company_slug(),
        ingest_type=IngestType.GIT,
        event_name=event_name,
        git_instance=git_config.instance_slug,
        git_provider=git_config.git_provider.value,
    )


class GitAdapter(ABC):
    config: GitConfig
    PULL_REQUEST_BATCH_SIZE_IN_BYTES = (
        50 * Constants.MB_SIZE_IN_BYTES
    )  # PRs can be huge and of variable size. We need to limit them by batch size in bytes
    NUMBER_OF_COMMITS_PER_BATCH = (
        30000  # Commits are generally uniform in size. This is ~50 MBs per commit batch
    )

    def _transform_data_objects_before_saving(
        self,
        dataclass_objects: (
            List[StandardizedObject]
            | List[StandardizedBranch]
            | List[StandardizedCommit]
            | List[StandardizedOrganization]
            | List[StandardizedPullRequest]
            | List[StandardizedRepository]
            | List[StandardizedTeam]
            | List[StandardizedUser]
            | List[StandardizedPullRequestAuthor]
            | List[StandardizedPullRequestReviewAuthor]
        ),
    ) -> List[Dict]:
        """Helper function for taking a list of objects that inherit from Dataclass and
        transforming them to a list of dictionary objects

        Args:
            dataclass_objects (List[DataclassInstance]): A list of Dataclass Instances

        Returns:
            List[Dict]: A list of dictionaries
        """

        def _transform(obj: StandardizedObject):
            if self.config.git_redact_names_and_urls:
                obj.redact_names_and_urls()
            if self.config.git_strip_text_content:
                obj.strip_text_content()
            return asdict(obj)

        return [_transform(dc_object) for dc_object in dataclass_objects]

    @staticmethod
    def get_git_adapter(config: GitConfig) -> "GitAdapter":
        """Static function for generating a GitAdapter from a provided GitConfig object

        Args:
            config (GitConfig): A git configuration data object. The specific GitAdapter
                is returned based on the git_provider field in this object

        Raises:
            GitProviderUnavailable: If the supplied git config has an unknown git provider, this error will be thrown

        Returns:
            GitAdapter: A specific subclass of the GitAdapter, based on what git_provider we need
        """
        from jf_ingest.jf_git.adapters.azure_devops import AzureDevopsAdapter
        from jf_ingest.jf_git.adapters.github import GithubAdapter
        from jf_ingest.jf_git.adapters.gitlab import GitlabAdapter

        if config.git_provider in [GitProviderInJellyfishRepo.GITHUB, GitProvider.GITHUB]:
            return GithubAdapter(config)
        elif config.git_provider in [GitProviderInJellyfishRepo.ADO, GitProvider.ADO]:
            return AzureDevopsAdapter(config)
        elif config.git_provider in [GitProviderInJellyfishRepo.GITLAB, GitProvider.GITLAB]:
            return GitlabAdapter(config)
        else:
            raise GitProviderUnavailable(
                f'Git provider {config.git_provider} is not currently supported'
            )

    @abstractmethod
    def get_api_scopes(self) -> str:
        """Return the list of API Scopes. This is useful for Validation

        Returns:
            str: A string of API scopes we have, given the adapters credentials
        """
        pass

    @abstractmethod
    def get_organizations(self) -> List[StandardizedOrganization]:
        """Get the list of organizations the adapter has access to

        Returns:
            List[StandardizedOrganization]: A list of standardized organizations within this Git Instance
        """
        pass

    @abstractmethod
    def get_users(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedUser, None, None]:
        """Get the list of users in a given Git Organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized Git Organization Object

        Returns:
            List[StandardizedUser]: A standardized User Object
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.
        """
        pass

    @abstractmethod
    def get_teams(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedTeam, None, None]:
        """Get the list of teams in a given Git Organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized Git Organization Object

        Returns:
            List[StandardizedUser]: A standardized Team Object
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.
        """
        pass

    @abstractmethod
    def get_repos(
        self,
        standardized_organization: StandardizedOrganization,
        limit: Optional[int] = None,
        only_private: bool = False,
    ) -> Generator[StandardizedRepository, None, None]:
        """Get a list of standardized repositories within a given organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized organization
            limit (int, optional): When provided, the number of items returned is limited.
            only_private (bool): When True, only private repositories will be returned. Defaults to False

        Returns:
            List[StandardizedRepository]: A list of standardized Repositories
        """
        pass

    @abstractmethod
    def get_repos_count(
        self, standardized_organization: StandardizedOrganization, only_private: bool = False
    ) -> int:
        """Get the count of repositories within a given organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized organization
            only_private (bool): When True, only private repositories will be counted. Defaults to False

        Returns:
            int: The count of repositories
        """
        pass

    @abstractmethod
    def get_commits_for_default_branch(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        """For a given repo, get all the commits that are on the Default Branch.

        Args:
            standardized_repo (StandardizedRepository): A standard Repository object
            limit (int): limit the number of commit objects we will yield
            pull_since (datetime): filter commits to be newer than this date
            pull_until (datetime): filter commits to be older than this date

        Returns:
            List[StandardizedCommit]: A list of standardized commits
        """
        pass

    @abstractmethod
    def get_branches_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_branches: Optional[bool] = False,
        limit: Optional[int] = None,
    ) -> Generator[StandardizedBranch, None, None]:
        """Function for pulling branches for a repository. By default, pull_branches will run as False,
        so we will only process the default branch. If pull_branches is true, than we will pull all
        branches in this repository

        Args:
            standardized_repo (StandardizedRepository): A standardized repo, which hold info about the default branch.
            pull_branches (bool): A boolean flag. If True, pull all branches available on Repo. If false, only process the default branch. Defaults to False.
            limit (int, optional): When provided, the number of items returned is limited.

        Yields:
            StandardizedBranch: A Standardized Branch Object
        """
        pass

    @abstractmethod
    def get_commits_for_branches(
        self,
        standardized_repo: StandardizedRepository,
        branches: List[StandardizedBranch],
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
        limit: Optional[int] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        """For a given repo, get all the commits that are on the included branches.
        Included branches are found by crawling across the branches pulled/available
        from get_filtered_branches

        Args:
            standardized_repo (StandardizedRepository): A standard Repository object
            branches (List[StandardizedBranch]): A list of branches to pull commits for
            pull_since (datetime): A date to pull from
            pull_until (datetime): A date to pull up to
            limit (Optional[int]): limit the number of commit objects we will yield

        Returns:
            List[StandardizedCommit]: A list of standardized commits
        """
        pass

    @abstractmethod
    def get_pr_metadata(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pr_pull_from_date: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequestMetadata, None, None]:
        """Get all PRs, but only included the bare necesaties

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.
            pr_pull_from_date: This is currently only used by the GithubAdapter. It is useful because
                the GithubAdapter caches Repository metadata, including when the Repo's most recent
                PR was updated. We can determine if we need to pull any PR data in memory, before seeing
                that value via the API

        Returns:
            List[StandardizedPullRequest]: A list of standardized PRs
        """
        pass

    @abstractmethod
    def git_provider_pr_endpoint_supports_date_filtering(self) -> bool:
        """Returns a boolean on if this PR supports time window filtering.
        So far, Github DOES NOT support this (it's adapter will return False)
        but ADO does support this (it's adapter will return True)

        Returns:
            bool: A boolean on if the adapter supports time filtering when searching for PRs
        """
        return False

    @abstractmethod
    def get_prs(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool = False,
        hash_files_for_prs: bool = False,
        limit: Optional[int] = None,
        start_cursor: Optional[Any] = None,
        start_window: Optional[datetime] = None,
        end_window: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequest, None, None]:
        """Get the list of standardized Pull Requests for a Standardized Repository.

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            pull_files_for_pr (bool): When provided, we will pull file metadata for all PRs
            hash_files_for_prs (bool): When provided, all file metadata will be hashed for PRs
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.

        Returns:
            List[StandardizedPullRequest]: A list of standardized PRs
        """
        pass

    @abstractmethod
    def get_pr_authors(
        self, pr_list: list[StandardizedJFAPIPullRequest]
    ) -> list[StandardizedPullRequestAuthor]:
        """Get the authors of a list of PRs

        This is currently only implemented for GitHub instances
        """
        pass

    @abstractmethod
    def get_pr_review_authors(
        self, pr_list: list[StandardizedJFAPIPullRequest]
    ) -> list[StandardizedPullRequestReviewAuthor]:
        """Get the review authors of a list of PRs

        This is currently only implemented for GitHub instances
        """
        pass

    def get_commits_for_repo(
        self, standardized_repo: StandardizedRepository, branches: List[StandardizedBranch]
    ) -> Generator[StandardizedCommit, None, None]:
        """This is a function that wraps the get_commits_for_branches function and applies the Repo
        backpopulation logic, if we need it to

        Args:
            standardized_repo (StandardizedRepository): A standardized Repository object
            branches (List[StandardizedBranches]): A list of branches to pull commits for

        Yields:
            Generator[StandardizedCommit, None, None]: A stream of commits. Potentially terminating early if we hit the pull from date
        """
        pull_from_for_commits = self.config.get_pull_from_for_commits()
        backpopulation_window = determine_commit_backpopulation_window(
            config=self.config, repo=standardized_repo
        )
        pull_until_for_commits = None

        # If backpopulating, set
        if backpopulation_window:
            backpopulation_start, backpopulation_end = backpopulation_window
            pull_from_for_commits = backpopulation_start
            pull_until_for_commits = backpopulation_end
            logging_helper.send_to_agent_log_file(
                f'Backpopulation was determeined as necessary for {standardized_repo.name}. Backpopulation will run from [{pull_from_for_commits}, {pull_until_for_commits}]'
            )
        else:
            logging_helper.send_to_agent_log_file(
                f'Backpopulation was not determined as necessary for {standardized_repo.name}. Commits will be pulled from {pull_from_for_commits}'
            )

        commit_count = 0
        commit = None
        for j, commit in enumerate(
            self.get_commits_for_branches(
                standardized_repo=standardized_repo,
                branches=branches,
                pull_since=pull_from_for_commits,
                pull_until=pull_until_for_commits,
            ),
            start=1,
        ):
            with logging_helper.log_loop_iters('branch commit inside repo', j, 100):
                # If we crawl across commits and find that we already have commits this old, stop processing
                # NOTE: THis is technically redundant, because the get_commits calls should have a pull_from/pull_until
                # scheme that should limit how many commits we pull
                if commit.commit_date and commit.commit_date < pull_from_for_commits:
                    break
                yield commit
                commit_count += 1
        if backpopulation_window:
            commits_backpopulated_to = None
            if commit:
                commits_backpopulated_to = max(
                    min(pull_from_for_commits, commit.commit_date), self.config.pull_from
                )
            else:
                commits_backpopulated_to = max(pull_from_for_commits, self.config.pull_from)
            standardized_repo.commits_backpopulated_to = commits_backpopulated_to
            logging_helper.send_to_agent_log_file(
                f'Setting commits_backpopulated_to for repo {standardized_repo.name} to {commits_backpopulated_to}'
            )
        logging_helper.send_to_agent_log_file(f'Found {commit_count} commits', level=logging.DEBUG)

    def get_prs_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool,
        hash_files_for_prs: bool,
    ) -> Generator[StandardizedPullRequest, None, None]:
        """This is a function that wraps the get_commits_for_branches function and applies the Repo
        backpopulation logic, if we need it to

        Args:
            standardized_repo (StandardizedRepository): A standardized Repository object

        Yields:
            Generator[StandardizedCommit, None, None]: A stream of commits. Potentially terminating early if we hit the pull from date
        """
        prs_start_cursor = None
        backpopulation_window = determine_pr_backpopulation_window(
            config=self.config, repo=standardized_repo
        )
        if backpopulation_window:
            pull_from_for_prs, pull_up_to_for_prs = backpopulation_window
            logging_helper.send_to_agent_log_file(
                f'Backpopulation was determined as necessary for Repo {standardized_repo.name} (ID: {standardized_repo.id}). '
                f'Backpop window: [{pull_from_for_prs}, {pull_up_to_for_prs}]'
            )
        else:
            pull_from_for_prs = self.config.get_pull_from_for_prs(standardized_repo.id)
            pull_up_to_for_prs = datetime.now().astimezone(timezone.utc) + timedelta(days=1)
            logging_helper.send_to_agent_log_file(
                f'Backpopulation was NOT determined as necessary for Repo {standardized_repo.name} (ID: {standardized_repo.id}). '
                f'PR Pull Window: [{pull_from_for_prs}, {pull_up_to_for_prs}]'
            )
        # If we are backpopulating and our Adapter DOES NOT support filtering for PRs with
        # datetime bounds, we need to find the starting mark of where to start
        # pulling PRs. To do this, we leverage the get_pr_metadata function, which should be
        # a light-weight alternative (in terms of API calls) to the get_prs function.
        # For an adapter that uses GQL, this alternative can be VERY light. For a non-GQL
        # adapter, this can be slightly lighter but likely not by much
        # NOTE: If a provider supports PR time filtering (like ADO), then this can be skipped!
        # It is faster to have the API do the filtering for us
        if backpopulation_window and not self.git_provider_pr_endpoint_supports_date_filtering():
            logging_helper.send_to_agent_log_file(
                f'Backpopulation window detected for PRs in {standardized_repo.name}, attempting to walk back on all PRs to find backpopulation window end date'
            )

            backpopulation_start, backpopulation_end = backpopulation_window
            pull_from_for_prs = backpopulation_start
            prs_found = False
            for api_pr_metadata in self.get_pr_metadata(
                standardized_repo=standardized_repo, pr_pull_from_date=pull_from_for_prs
            ):
                if api_pr_metadata.updated_at > backpopulation_end:
                    logging_helper.send_to_agent_log_file(
                        f'Backpopulation flow -- skipping PR (ID: {api_pr_metadata.id}) from {api_pr_metadata.updated_at} '
                        f'because backpopulation_end is {backpopulation_end.isoformat()} (Repo: {standardized_repo.name})'
                    )
                    # This is the START cursor, so it is NON-INCLUSIVE. We want it to be trailing by 1 index
                    prs_start_cursor = api_pr_metadata.api_index
                    continue
                else:
                    if api_pr_metadata.updated_at <= self.config.pull_from:
                        logging_helper.send_to_agent_log_file(
                            f'Exiting backpopulation walkback loop and NOT ingesting this PR, because PR {api_pr_metadata.id} was last updated at {api_pr_metadata.updated_at} which is less than our base pull from date: {self.config.pull_from}'
                        )
                        standardized_repo.prs_backpopulated_to = self.config.pull_from
                        return
                    elif api_pr_metadata.updated_at <= backpopulation_start:
                        # We want to ingest this one PR in this case, because it will greatly fast forward our backpopulation dates
                        logging_helper.send_to_agent_log_file(
                            f'Exiting backpopulation walkback loop, because PR {api_pr_metadata.id} was last updated at {api_pr_metadata.updated_at} which is less than our backpopulation start time ({backpopulation_start}). We WILL ingest this PR'
                        )
                        prs_found = True
                        break
                    else:
                        logging_helper.send_to_agent_log_file(
                            f'Exiting backpopulation walkback loop, because PR {api_pr_metadata.id} was last updated at {api_pr_metadata.updated_at} which is within our backpopulation window ([{backpopulation_start}, {backpopulation_end}]). We will ingest this PR and all other PRs up until {backpopulation_start}'
                        )
                        prs_found = True
                        break

            if not prs_found:
                logging_helper.send_to_agent_log_file(
                    f'No PRs found when looking in and beyond our backpopulation window, setting PRs backpopulated to to the pull from date for this git instance {self.config.pull_from}'
                )
                standardized_repo.prs_backpopulated_to = self.config.pull_from
                return

        pr = None
        pr_count_for_repo = 0
        get_pr_runs = 0
        for i, pr in enumerate(
            self.get_prs(
                standardized_repo=standardized_repo,
                pull_files_for_pr=pull_files_for_pr,
                hash_files_for_prs=hash_files_for_prs,
                start_cursor=prs_start_cursor,
                start_window=pull_from_for_prs,
                end_window=pull_up_to_for_prs,
            ),
            start=1,
        ):
            get_pr_runs += 1
            with logging_helper.log_loop_iters('pr inside repo', i, 10):
                # If we crawl across prs and find that we already have PR this old, stop processing
                if (
                    not self.git_provider_pr_endpoint_supports_date_filtering()
                    and pr.updated_at
                    and pr.updated_at <= pull_from_for_prs
                ):
                    logging_helper.send_to_agent_log_file(
                        f'Stopping PR crawl for repo {standardized_repo.name} because PR {pr.id} as been identified as being older than the pull from date ({pr.updated_at} <= {pull_from_for_prs}).'
                    )
                    # If we're backpopulating, this PR represents the next oldest PR. If we ingest it, we can speed up
                    # the backpopulation window to be as old as this PR. Only ingest it if it's within the parent 'pull_from' window, though
                    if backpopulation_window and pr.updated_at >= self.config.pull_from:
                        logging_helper.send_to_agent_log_file(
                            f'This PR ({pr.id}) will be ingest by Jellyfish, because are backpopulating this repo ({standardized_repo.name})'
                        )
                        yield pr
                        pr_count_for_repo += 1
                    break
                yield pr
                pr_count_for_repo += 1
        add_telemetry_fields({'get_pr_runs': get_pr_runs})
        # If we're backpopulating, update the prs_back_populated_to variable
        if backpopulation_window:
            prs_back_populated_to = None
            if pr:
                prs_back_populated_to = max(
                    min(pull_from_for_prs, pr.updated_at), self.config.pull_from
                )
            else:
                prs_back_populated_to = max(pull_from_for_prs, self.config.pull_from)
            standardized_repo.prs_backpopulated_to = prs_back_populated_to
            logging_helper.send_to_agent_log_file(
                f'Setting prs_backpopulated_to for repo {standardized_repo.name} to {prs_back_populated_to}'
            )

        logging_helper.send_to_agent_log_file(
            f'{pr_count_for_repo} PRs found for repo {standardized_repo.name}'
        )

    def get_filtered_branches(
        self, repo: StandardizedRepository, branches: List[StandardizedBranch]
    ) -> set[str]:
        """Return branches for which we should pull commits, specified by customer in git config.
            The repo's default branch will always be included in the returned list.

        Args:
            repo (StandardizedRepository): A standardized repository

        Returns:
            set[str]: A set of branch names (as strings)
        """

        # Helper function
        def get_matching_branches(
            included_branch_patterns: List[str], repo_branch_names: List[Optional[str]]
        ) -> List[str]:
            # Given a list of patterns, either literal branch names or names with wildcards (*) meant to match a set of branches in a repo,
            # return the list of branches from repo_branches that match any of the branch name patterns.
            # fnmatch is used over regex to support wildcards but avoid complicating the requirements on branch naming in a user's config.
            matching_branches = []
            for repo_branch_name in repo_branch_names:
                if not repo_branch_name:
                    continue
                elif self.config.pull_all_commits_and_branches:
                    matching_branches.append(repo_branch_name)
                elif any(
                    fnmatch.fnmatch(repo_branch_name, pattern)
                    for pattern in included_branch_patterns
                ):
                    matching_branches.append(repo_branch_name)
            return matching_branches

        # Always process the default branch
        branches_to_process = [repo.default_branch_name] if repo.default_branch_name else []
        # Agent use case: check for the included_branches values
        additional_branches_for_repo: List[str] = self.config.included_branches_by_repo.get(
            repo.name, []
        )

        # Extend and potentially filter branches to process
        repo_branch_names = [b.name for b in branches if b]
        branches_to_process.extend(
            get_matching_branches(additional_branches_for_repo, repo_branch_names)
        )
        return set(branches_to_process)

    def discover_new_orgs(self) -> List[StandardizedOrganization]:
        """Helper function for discovering new Git organizations. Currently
        not implemented because only Github has been implement as a GitAdapter
        subclass, and Github DOES NOT support discovering new orgs. Orgs must
        be entered manually

        Raises:
            NotImplementedError: Error stating that this function is net yet to use yet

        Returns:
            List[StandardizedOrganization]: A list of standardized Org Objects
        """
        raise NotImplementedError('Discover New Orgs is not yet implemented')

    def load_and_dump_git(self, git_config: GitConfig, ingest_config: IngestionConfig):
        """This is a shared class function that can get called by
        the different types of GitAdapters that extend this class.
        This function handles fetching all the necessary data from
        Git, transforming it, and saving it to local disk and/or S3

        Args:
            ingest_config (IngestionConfig): A valid Ingestion Config
        """
        init_jf_ingest_run(ingestion_config=ingest_config)
        self._run_load_and_dump_git(git_config=git_config, ingest_config=ingest_config)

    def _run_load_and_dump_git(self, git_config: GitConfig, ingest_config: IngestionConfig):
        #######################################################################
        # Init IO Helper
        #######################################################################
        ingest_io_helper = IngestIOHelper(ingest_config=ingest_config)

        # Wrapper function for writing to the IngestIOHelper
        def _write_to_s3_or_local(object_name: str, json_data: list[dict], batch_number: int = 0):
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=object_name,
                json_data=json_data,
                subdirectory=SubDirectory.GIT,
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
                git_instance_key=self.config.instance_file_key,
                batch_number=batch_number,
            )

        #######################################################################
        # ORGANIZATION DATA
        #######################################################################
        with record_span('get_organizations'):
            logger.info('Fetching Git Organization Data...')
            try:
                standardized_organizations: List[StandardizedOrganization] = (
                    self.get_organizations()
                )
            except HTTPError as e:
                resp = getattr(e, 'response', None)
                if resp and resp.status_code in [401, 403, 404]:
                    raise GitAuthorizationException(
                        f'Unable to fetch organizations for {git_config.git_provider} {git_config.instance_slug}: {e}'
                    )
                raise
            logger.info(
                f'Successfully pulled Git Organizations data for {len(standardized_organizations)} Organizations.'
            )
            add_telemetry_fields({'git_organization_count': len(standardized_organizations)})
        # Upload Data
        _write_to_s3_or_local(
            object_name=GitObject.GitOrganizations.value,
            json_data=self._transform_data_objects_before_saving(standardized_organizations),
        )

        #######################################################################
        # USER DATA
        #######################################################################
        if not git_config.skip_pulling_users:
            logger.info('Fetching Git User Data...')
            with record_span('get_users'):
                standardized_users: List[StandardizedUser] = [
                    user
                    for org in standardized_organizations
                    for user in tqdm_to_logger(
                        self.get_users(org), desc='Processing Users', unit=' users'
                    )
                ]
                add_telemetry_fields({'git_user_count': len(standardized_users)})
            logger.info(f'Successfully found {len(standardized_users)} users.')
            # Upload Data
            _write_to_s3_or_local(
                object_name=GitObject.GitUsers.value,
                json_data=self._transform_data_objects_before_saving(standardized_users),
            )
        else:
            _write_to_s3_or_local(
                object_name=GitObject.GitUsers.value,
                json_data=[],
            )
            logger.info(
                f'Not pulling users because \'skip_pulling_users\' is set to: {git_config.skip_pulling_users}.'
            )

        #######################################################################
        # TEAM DATA
        #######################################################################
        with record_span('get_teams'):
            logger.info('Fetching Git Team Data...')
            standardized_teams: List[StandardizedTeam] = (
                [
                    team
                    for org in standardized_organizations
                    for team in tqdm_to_logger(
                        self.get_teams(org), desc="Processing Teams", unit=" teams"
                    )
                ]
                if self.config.pull_teams
                else []
            )
            logger.info(f'Successfully found {len(standardized_teams)} teams.')
            add_telemetry_fields({'git_team_count': len(standardized_teams)})
        # Upload Data
        _write_to_s3_or_local(
            object_name=GitObject.GitTeams.value,
            json_data=self._transform_data_objects_before_saving(standardized_teams),
        )

        #######################################################################
        # REPO DATA, NOTE THAT WE UPLOAD LATER BECAUSE WE NEED TO SET
        # THE BACKPOPULATD DATES BELOW AFTER WE PULL PRS AND COMMITS
        #######################################################################
        if not git_config.skip_pulling_repos:
            with record_span('get_repos'):
                logger.info('Fetching Git Repo Data...')

                # Gitlab orgs can share repos. To prevent duplicates, track the ids seen.
                repo_ids_seen: Set[str] = set()
                standardized_repos: List[StandardizedRepository] = []
                for org in standardized_organizations:
                    with record_span(
                        'get_repos_for_org',
                        {'org_name': str(org.name), 'org_login': str(org.login)},
                    ):
                        for repo in tqdm_to_logger(
                            self.get_repos(
                                standardized_organization=org,
                            ),
                            unit=' repos',
                            desc=f'Pulling all available Repositories',
                        ):
                            repo_id = repo.id
                            if repo_id in repo_ids_seen:
                                continue
                            standardized_repos.append(repo)
                            repo_ids_seen.add(repo_id)

                logger.info(
                    f'Successfully pulled Git Repo Data for {len(standardized_repos)} Repos.'
                )
                add_telemetry_fields({'git_repo_count': len(standardized_repos)})
        else:
            logger.info(
                f'Not pulling new repo data because \'skip_pulling_repos\' is set to {git_config.skip_pulling_repos}. '
                f'We will pull data for the {len(git_config.repos_in_jellyfish)} repos that already exist in Jellyfish'
            )
            standardized_repos = git_config.repos_in_jellyfish

        repos_to_process = [
            repo for repo in standardized_repos if repo.id not in git_config.quiescent_repos
        ]

        filters = []
        if self.config.included_repos:
            logger.info(f'Filtering repos to only include {self.config.included_repos}')

            def check_included_repo(repo_to_check: StandardizedRepository) -> bool:
                for repo_to_include in self.config.included_repos:
                    if repo_to_check.name.lower() == str(repo_to_include).lower() or str(
                        repo_to_check.id
                    ) == str(repo_to_include):
                        return True
                return False

            filters.append(check_included_repo)
        if self.config.excluded_repos:
            logger.info(f'Filtering repos to exclude {self.config.excluded_repos}')

            def check_excluded_repo(repo_to_check: StandardizedRepository) -> bool:
                for repo_to_exclude in self.config.excluded_repos:
                    if repo_to_check.name.lower() == str(repo_to_exclude).lower() or str(
                        repo_to_check.id
                    ) == str(repo_to_exclude):
                        return False
                return True

            filters.append(check_excluded_repo)

        repos_to_process = [
            repo for repo in repos_to_process if all(filt(repo) for filt in filters)
        ]

        repo_count = len(repos_to_process)

        logging_helper.send_to_agent_log_file(
            f'Processing {len(repos_to_process)}. {len(git_config.quiescent_repos)} were marked as being quiescent'
        )

        #######################################################################
        # BRANCH DATA
        # NOTE: Branches are optionally processed, depending on GitConfiguration.
        # For Direct Connect it is likely we only process the default branch,
        # for agent we process all branches
        #######################################################################
        repo_to_branches: dict[str, List[StandardizedBranch]] = {}
        all_branches = []
        with tqdm_to_logger(desc='Processing Branches', unit=' Branches') as pbar:
            with record_span('get_branches_for_repos'):
                for repo in repos_to_process:
                    pull_branches = (
                        git_config.pull_all_commits_and_branches
                        or git_config.repo_id_to_pull_all_commits_and_branches.get(repo.id)
                    )
                    branch_batch = []
                    # Iterate across branches and update the progress bar so we can see
                    # counts and rates of branch processing
                    for branch in self.get_branches_for_repo(repo, pull_branches):
                        pbar.update(1)
                        branch_batch.append(branch)

                    repo_to_branches[repo.id] = branch_batch
                    all_branches.extend(branch_batch)
                add_telemetry_fields({'git_branch_count': len(all_branches)})

        _write_to_s3_or_local(
            object_name=GitObject.GitBranches.value,
            json_data=self._transform_data_objects_before_saving(all_branches),
        )

        #######################################################################
        # COMMIT DATA
        #
        # NOTE: Commit data can be quite large, so for better memory handling
        # we will create a chain of generators (get_commits_for_branches returns a generator)
        # and process our way through those generators, uploading data ~50 MBs at a time
        # NOTE: Commit data is pretty uniform in size (each commit is ~2KB), so we'll upload
        # in batches of 30k commits (roughly 50 MB in data)
        #
        #######################################################################
        total_commits = 0
        logger.info(f'Fetching Git Commit Data for {repo_count} Repos...')
        list_of_commit_generators: List[Generator[StandardizedCommit, None, None]] = []
        with record_span('get_commits_for_repos'):
            for repo in repos_to_process:
                branches = repo_to_branches[repo.id]
                commit_generator_for_repo = self.get_commits_for_repo(repo, branches=branches)
                list_of_commit_generators.append(commit_generator_for_repo)

            # Chain together all the generators
            commits_generator = tqdm_to_logger(
                chain.from_iterable(list_of_commit_generators),
                desc=f'Processing Commits for {repo_count} repos',
                unit=' commits',
            )
            for batch_num, commit_batch in enumerate(
                batch_iterable(commits_generator, batch_size=self.NUMBER_OF_COMMITS_PER_BATCH)
            ):
                total_commits += len(commit_batch)
                commit_batch_as_dict = self._transform_data_objects_before_saving(commit_batch)
                _write_to_s3_or_local(
                    object_name=GitObject.GitCommits.value,
                    json_data=commit_batch_as_dict,
                    batch_number=batch_num,
                )
            if not total_commits:
                _write_to_s3_or_local(object_name=GitObject.GitCommits.value, json_data=[])
            logger.info(f'Successfully process {total_commits} total commits')
            add_telemetry_fields({'git_commit_count': total_commits})

        #######################################################################
        # PULL REQUEST DATA
        #
        # NOTE: Pull Request data can be quite large, so for better memory handling
        # we will create a chain of generators (get_prs returns a generator)
        # and process our way through those generators, uploading data ~50 MBs at a time
        #
        #######################################################################
        total_prs = 0
        logger.info(f'Fetching Git Pull Request Data for {repo_count} Repos...')
        list_of_pr_generators: List[Generator[StandardizedPullRequest, None, None]] = []
        with record_span('get_prs_for_repos'):
            for repo in repos_to_process:
                if self.config.repos_to_skip_pull_prs_for.get(repo.id):
                    logging_helper.send_to_agent_log_file(
                        f'Skipping pull PRs for {repo.name} because it was marked in Jellyfish as so'
                    )
                    continue
                pr_generator_for_repo = self.get_prs_for_repo(
                    repo,
                    pull_files_for_pr=git_config.pull_files_for_prs,
                    hash_files_for_prs=git_config.hash_files_for_prs,
                )
                list_of_pr_generators.append(pr_generator_for_repo)

            # Chain together all the generators
            prs_generator = tqdm_to_logger(
                chain.from_iterable(list_of_pr_generators),
                desc=f'Processing Pull Request Data for {repo_count} repos',
                unit=' PRs',
            )
            for batch_num, pr_batch in enumerate(
                batch_iterable_by_bytes_size(
                    prs_generator, batch_byte_size=self.PULL_REQUEST_BATCH_SIZE_IN_BYTES
                )
            ):
                total_prs += len(pr_batch)
                pr_batch_as_dict = self._transform_data_objects_before_saving(pr_batch)
                _write_to_s3_or_local(
                    object_name=GitObject.GitPullRequests.value,
                    json_data=pr_batch_as_dict,
                    batch_number=batch_num,
                )

            if not total_prs:
                # IF we don't have any PRs, push an empty file
                _write_to_s3_or_local(
                    object_name=GitObject.GitPullRequests.value,
                    json_data=[],
                    batch_number=0,
                )

            logger.info(f'Successfully processed {total_prs} total PRs')
            add_telemetry_fields({'git_pr_count': total_prs})

        # For GHC instances, make a request to the JF API to get any PRs that have mannequin authors with
        # missing user data (or missing review authors)
        if (
            git_config.git_provider in [GitProviderInJellyfishRepo.GITHUB, GitProvider.GITHUB]
            and git_config.check_ghc_mannequin_user_prs
        ):
            logger.info(
                f'Checking for any JF PRs with missing mannequin user data for instance {git_config.instance_slug}'
            )
            prs_with_null_user_data = get_jf_github_null_user_pr_data(ingest_config, git_config)
            prs_with_null_review_data = get_jf_github_null_user_pr_data(
                ingest_config, git_config, pr_reviews=True
            )

            if prs_with_null_user_data:
                logger.info(
                    f'JF returned {len(prs_with_null_user_data)} PRs with missing mannequin user data'
                )

                pr_authors = self.get_pr_authors(prs_with_null_user_data)
                _write_to_s3_or_local(
                    object_name=GitObject.GitPullRequestAuthors.value,
                    json_data=self._transform_data_objects_before_saving(pr_authors),
                )

                logger.info(f'Wrote {len(pr_authors)} PR authors to S3/local')

            if prs_with_null_review_data:
                logger.info(
                    f'JF returned {len(prs_with_null_review_data)} PRs with missing mannequin review data'
                )

                pr_review_authors = self.get_pr_review_authors(prs_with_null_review_data)
                _write_to_s3_or_local(
                    object_name=GitObject.GitPullRequestReviewAuthors.value,
                    json_data=self._transform_data_objects_before_saving(pr_review_authors),
                )

                logger.info(f'Wrote {len(pr_review_authors)} PR review authors to S3/local')
        else:
            logger.debug(
                f'Skipping GHC mannequin user PRs check ({git_config.instance_slug}, {git_config.check_ghc_mannequin_user_prs}, {git_config.git_provider})'
            )

        # Upload Repo Data at the very end
        _write_to_s3_or_local(
            object_name=GitObject.GitRepositories.value,
            json_data=(
                self._transform_data_objects_before_saving(repos_to_process)
                if repos_to_process
                else []
            ),
        )

        logger.info(f'Done processing Git Data!')

    @logging_helper.log_entry_exit()
    def validate_instance_authorization(self) -> None:
        """
        Validate our authorization to the Git instance is sufficient for collecting
        the necessary data.

        Raises:
            GitAuthorizationException: Raised when unable to access a required resource
        """
        if self.config.git_provider == GitProviderInJellyfishRepo.GITHUB_ENTERPRISE_CLOUD:
            logger.info('Skipping authorization check for Github Enterprise Cloud instance')
            return None

        logger.info(
            f'Verifying authorization to {self.config.company_slug} {self.config.git_provider.name} instance...'
        )

        # If discover orgs is set to false, we'll track instances where we can successfully make requests with
        # the authentication provided, but are unable to receive any results
        discover_orgs = self.config.discover_organizations
        logger.info(f'Discover organizations set to {discover_orgs}')

        # Some Github instances are only used for copilot - in this case, it's expected that we won't be able to
        # pull any data. In the future, we'll have a flag to denote whether the instance is *only* used for copilot,
        # but for now, we'll just check if the instance has it enabled
        copilot_enabled = self.config.copilot_enabled
        logger.info(f'Copilot enabled set to {copilot_enabled}')

        # Providing a pull since datetime is required for Github when pulling commits. Setting it to an old date to
        # account for old repositories that may have been inactive for some time
        pull_since_date = datetime(1970, 1, 1, tzinfo=timezone.utc)

        if discover_orgs:
            try:
                orgs = self.get_organizations()
                logger.info(f'Discovered {len(orgs)} accessible organizations')
            except Exception as e:
                raise GitAuthorizationException(
                    f'Unable to access organizations for {self.config.git_provider}: {e}'
                )
        else:
            logger.info(
                f'Using {len(self.config.git_organizations)} provided organizations from JF DB'
            )
            orgs = [
                StandardizedOrganization(id=org, name=None, login=org, url=None)
                for org in self.config.git_organizations
            ]

        # Keep track of auth errors rather than raising immediately to capture all potential issues
        auth_errors: dict[str, list[str]] = {}

        # Some Gitlab instances can have thousands of orgs (aka groups), which will make the auth validator
        # take a long time to run. To prevent this, we'll break the look if 10 orgs have been successfully checked
        is_gitlab_provider = bool(self.config.git_provider == GitProviderInJellyfishRepo.GITLAB)
        successful_org_auth_count = 0

        for idx, org in enumerate(orgs, start=1):
            if org.login in self.config.excluded_organizations:
                logger.info(f'Skipping authorization check for excluded org {org.login}')
                continue

            logger.info(f'Verifying authorization to org {org.login} ({idx}/{len(orgs)})')
            any_resource_accessed = False
            resource_accessed_types: list[str] = []
            org_auth_errors: list[str] = []

            if not self.config.skip_pulling_users:
                if _validate_resource_access(
                    resource_name='users',
                    fetch_func=self.get_users,
                    fetch_kwargs={'standardized_organization': org, 'limit': 10},
                    org_auth_errors=org_auth_errors,
                    discover_orgs=discover_orgs,
                    record_empty=True,
                ):
                    any_resource_accessed = True
                    resource_accessed_types.append('users')

            # Gitlab does not have the concept of teams other than groups, which we classify as orgs
            if (
                not self.config.git_provider == GitProviderInJellyfishRepo.GITLAB
                and self.config.pull_teams
            ):
                if _validate_resource_access(
                    resource_name='teams',
                    fetch_func=self.get_teams,
                    fetch_kwargs={'standardized_organization': org, 'limit': 10},
                    org_auth_errors=org_auth_errors,
                    discover_orgs=discover_orgs,
                    record_empty=True,
                ):
                    any_resource_accessed = True
                    resource_accessed_types.append('teams')

            if not self.config.skip_pulling_repos:
                repos, get_repos_err = self._get_repos_for_auth_check(org)
                repo_data_access = False

                # If we encountered an error while trying to access repositories, the returned list of repos will
                # be empty, and this block will be skipped
                if repos:
                    branch_access = False
                    commit_access = False
                    pr_access = False

                    for repo_idx, repo in enumerate(repos, start=1):
                        logger.info(
                            f'Attempting auth validation on repo {repo.name} (attempt {repo_idx}/{len(repos)})'
                        )

                        if branches := _validate_resource_access(
                            resource_name='branches',
                            fetch_func=self.get_branches_for_repo,
                            fetch_kwargs={
                                'standardized_repo': repo,
                                'pull_branches': True,
                                'limit': 10,
                            },
                            org_auth_errors=org_auth_errors,
                            discover_orgs=discover_orgs,
                            record_empty=False,
                        ):
                            branch_access = True
                            resource_accessed_types.append('branches')

                            if _validate_resource_access(
                                resource_name='commits',
                                fetch_func=self.get_commits_for_branches,
                                fetch_kwargs={
                                    'standardized_repo': repo,
                                    'branches': branches,
                                    'pull_since': pull_since_date,
                                    'limit': 1,
                                },
                                org_auth_errors=org_auth_errors,
                                discover_orgs=discover_orgs,
                                record_empty=False,
                            ):
                                commit_access = True
                                resource_accessed_types.append('commits')

                        if _validate_resource_access(
                            resource_name='prs',
                            fetch_func=self.get_prs,
                            fetch_kwargs={
                                'standardized_repo': repo,
                                'pull_files_for_pr': True,
                                'hash_files_for_prs': False,
                                'limit': 1,
                            },
                            org_auth_errors=org_auth_errors,
                            discover_orgs=discover_orgs,
                            record_empty=False,
                        ):
                            pr_access = True
                            resource_accessed_types.append('prs')

                        if any([branch_access, commit_access, pr_access]):
                            any_resource_accessed = True
                            repo_data_access = True

                        # Only break if we've accessed all types of repository resources
                        if all([branch_access, commit_access, pr_access]):
                            break
                elif copilot_enabled and not get_repos_err:
                    logger.info(
                        f'No repositories found, but copilot is enabled - skipping repo auth check for org {org.login}'
                    )
                    continue
                else:
                    org_auth_errors.append('Unable to access any organization repositories')

                if not get_repos_err and repos:
                    if not repo_data_access and not discover_orgs:
                        org_auth_errors.append(
                            f'Unable to access branches, commits or PRs from any of the first {len(repos)} accessible repos'
                        )
                    else:
                        # If we have discover_orgs set to True, only check that we didn't raise any exceptions
                        logger.info(f'Authorization validated for repo data in {org.login}')
                elif get_repos_err:
                    org_auth_errors.append(get_repos_err)

            if org_auth_errors:
                msg_prefix = (
                    f'Found {len(org_auth_errors)} authorization error(s) in org {org.login}'
                )
                exc_raised = any('raised' in err for err in org_auth_errors)

                if any_resource_accessed and not exc_raised:
                    logger.warning(f'{msg_prefix}, but accessed: {resource_accessed_types}')

                    # If any resources were accessed and no exception was raised, we can count that as
                    # the authorization check passing
                    successful_org_auth_count += 1
                elif not exc_raised:
                    resource_err_msg = (
                        'All resource types were accessible, but returned zero results'
                    )
                    logger.error(f'{msg_prefix}, {resource_err_msg}')
                    auth_errors[org.login] = [resource_err_msg]
                else:
                    logger.error(f'{msg_prefix}, including a raised exception')
                    auth_errors[org.login] = org_auth_errors
            else:
                logger.info(
                    f'Completed authorization validation for all resources in org {org.login}'
                )
                successful_org_auth_count += 1

            if successful_org_auth_count >= 10 and is_gitlab_provider:
                logger.info(
                    'Reached 10 successfully authorized organizations, stopping authorization check on Gitlab instance'
                )
                break

        if auth_errors:
            error_message = (
                f'Found {len(auth_errors)}/{len(orgs)} organizations with authorization errors:\n'
            )

            for org_name, org_auth_errors in auth_errors.items():
                error_message += f'Organization: {org_name}\n'
                for error in org_auth_errors:
                    error_message += f' - {error}\n'

            raise GitAuthorizationException(error_message)
        else:
            logger.info(
                f'Successfully authorized access to all necessary resources in {self.config.git_provider.value} instance for {self.config.company_slug}'
            )

    def _get_repos_for_auth_check(
        self, org: StandardizedOrganization, limit: int = 10
    ) -> tuple[list[StandardizedRepository], Optional[str]]:
        """
        Get repositories for authorization check. If the number of private repositories is fewer than the limit,
        we will supplement the list with public repositories in an attempt to reach the limit. If an exception
        is raised while trying to access repositories, the exception message will be returned.

        Args:
            org (StandardizedOrganization): The organization to get repositories for
            limit (int, optional): The maximum number of repositories to return. Defaults to 10.

        Returns:
            tuple[list[StandardizedRepository], bool]: A tuple containing a list of repositories and a string
                                                       if an exception was raised, None otherwise
        """
        repos: List[StandardizedRepository] = []
        exception_raised: Optional[str] = None

        try:
            # Check if the org has any private repositories. If the provider is ADO, we don't need to
            # check for private (aka hidden) repos, since we don't normally pull this data
            private_repo_count = (
                bool(self.get_repos_count(org, only_private=True))
                if self.config.git_provider != GitProvider.ADO
                else False
            )

            if private_repo_count:
                logger.info(f'Attempting to access private repos for org {org.login}')
                priv_repos = [
                    r
                    for r in self.get_repos(org, limit=limit, only_private=True)
                    if r.id not in self.config.excluded_repos
                ]
                repos.extend(priv_repos)
            else:
                logger.warning('No private repositories found')

            if len(repos) < limit:
                logger.warning(
                    f'{len(repos)} private repos found, attempting to access public repos for org {org.login}'
                )
                public_repos = [
                    r
                    for r in self.get_repos(org, limit=limit - len(repos), only_private=False)
                    if r.id not in self.config.excluded_repos
                ]
                repos.extend(public_repos)
        except Exception as e:
            exception_raised = f'Unable to access repositories in org {org.login}: {str(e)}'
            logger.error(exception_raised)

        if not exception_raised:
            logger.info(f'Found {len(repos)} repositories for org {org.login}')

        return repos, exception_raised


@jelly_trace
@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def load_and_push_git_to_s3(ingest_config: IngestionConfig) -> list[Exception]:
    """Handler function for the end to end processing of Git Data.
    This function is responsible for taking in an ingest config,
    creating a git adapter, and then running the Git Adapter function
    for uploading data to S3 (or saving it locally). The function for
    handling that logic is part of the GitAdapter class (see load_and_dump_git)

    Args:
        ingest_config (IngestionConfig): A fully formed IngestionConfig class, with
        valid Git Configuration in it.
    """
    exceptions: list[Exception] = []

    for git_config in ingest_config.git_configs:
        try:
            add_telemetry_fields({'company_slug': ingest_config.company_slug})
            git_adapter: GitAdapter = GitAdapter.get_git_adapter(git_config)
            with record_span('load_and_dump_git'):
                add_telemetry_fields(
                    {
                        'git_provider': git_config.git_provider.value,
                        'instance_slug': git_config.instance_slug,
                    }
                )
                git_adapter.load_and_dump_git(git_config=git_config, ingest_config=ingest_config)
        except GitProviderUnavailable:
            logger.warning(
                f'Git Config for provider {git_config.git_provider} is currently NOT supported!'
            )
            exceptions.append(JFIngestGitProviderException(git_config.git_provider))
            continue
        except Exception as e:
            logger.error(f'Error processing Git Data for {git_config.instance_slug}')
            logging_helper.send_to_agent_log_file(traceback.format_exc(), level=logging.ERROR)
            exceptions.append(e)

    return exceptions


def determine_commit_backpopulation_window(
    config: GitConfig, repo: StandardizedRepository
) -> Optional[BackpopulationWindow]:
    """Get the backpopulation window for Commits

    Args:
        config (GitConfig): A valid Git Config
        repo (StandardizedRepository): A valid standardized repository

    Returns:
        BackpopulationWindow: A Backpopulation window object
    """
    commits_backpopulated_to = config.get_backpopulated_date_for_commits(repo.id)
    return _get_backpopulation_helper(
        repo=repo,
        pull_from=config.pull_from,
        objects_back_populated_to=commits_backpopulated_to,
        object_name='commits',
        force_full_backpopulation_pull=config.force_full_backpopulation_pull,
        backpopulation_window_days=config.backpopulation_window_days,
    )


def determine_pr_backpopulation_window(
    config: GitConfig, repo: StandardizedRepository
) -> Optional[BackpopulationWindow]:
    """Get the backpopulation window for PRs

    Args:
        config (GitConfig): A valid Git Config
        repo (StandardizedRepository): A valid standardized repository

    Returns:
        BackpopulationWindow: A Backpopulation window object
    """
    prs_backpopulated_to = config.get_backpopulated_date_for_prs(repo.id)
    return _get_backpopulation_helper(
        repo=repo,
        pull_from=config.pull_from,
        objects_back_populated_to=prs_backpopulated_to,
        object_name='PRs',
        force_full_backpopulation_pull=config.force_full_backpopulation_pull,
        backpopulation_window_days=config.backpopulation_window_days,
    )


def _get_backpopulation_helper(
    repo: StandardizedRepository,
    pull_from: datetime,
    objects_back_populated_to: Optional[datetime],
    object_name: str,
    force_full_backpopulation_pull: bool = False,
    backpopulation_window_days: int = 30,
) -> Optional[BackpopulationWindow]:
    if objects_back_populated_to and objects_back_populated_to <= pull_from:
        # No backpopulation necessary
        return None
    # We're backpopulating objects for this repo

    if objects_back_populated_to:
        base_date = objects_back_populated_to
    else:
        base_date = datetime.now().astimezone(timezone.utc) + timedelta(days=1)

    backpopulation_window_start = (
        pull_from
        if force_full_backpopulation_pull
        else max(pull_from, base_date - timedelta(days=backpopulation_window_days))
    )
    backpopulation_window_end = base_date

    logging_helper.send_to_agent_log_file(
        f'Backpopulation window found for {object_name} for repo {repo.name} (ID: {repo.id}). Window spans from {backpopulation_window_start} to {backpopulation_window_end} ({object_name} backpopulated to {objects_back_populated_to}, pull_from: {pull_from})'
    )
    return BackpopulationWindow(backpopulation_window_start, backpopulation_window_end)


def _validate_resource_access(
    resource_name: str,
    fetch_func: Callable[..., Iterable[Any]],
    fetch_kwargs: dict[str, Any],
    org_auth_errors: list[str],
    discover_orgs: bool,
    record_empty: bool,
) -> list[Any]:
    results: list[Any] = []

    try:
        results = [i for i in fetch_func(**fetch_kwargs)]

        if results:
            logger.info(f"Successfully accessed {resource_name}")
        else:
            empty_msg = f"No {resource_name} found"
            logger.warning(empty_msg)

            if not discover_orgs and record_empty:
                org_auth_errors.append(empty_msg)
    except Exception as e:
        error_msg = f"Exception raised when attempting to access {resource_name}: {e}"
        logger.error(error_msg)
        org_auth_errors.append(error_msg)

    return results



import json

import requests_mock

from jf_ingest.jf_jira.downloaders import download_issuelinktypes
from tests.jf_jira.utils import (
    _register_jira_uri,
    get_fixture_file_data,
    get_jira_mock_connection,
)


def test_download_issuelinktypes():
    """
    Resolutions is a simple function, so we just need a simple test
    """
    with requests_mock.Mocker() as m:
        issuelinktype_data_str = get_fixture_file_data("api_responses/issueLinkType.json")
        _register_jira_uri(m, endpoint="issueLinkType", return_value=issuelinktype_data_str)

        issuelinktypes = download_issuelinktypes(get_jira_mock_connection())

        assert type(issuelinktypes) == list
        assert issuelinktypes == json.loads(issuelinktype_data_str)["issueLinkTypes"]



"""
Integration tests for JiraCloudManifestAdapter API selection functionality.
"""
import unittest
from unittest.mock import MagicMock, patch
from types import MappingProxyType

from jf_ingest.config import JiraDownloadConfig
from jf_ingest.constants import Constants
from jf_ingest.data_manifests.jira.adapters.jira_cloud import JiraCloudManifestAdapter


class TestJiraCloudManifestAdapterAPISelection(unittest.TestCase):
    """Test API selection functionality in JiraCloudManifestAdapter."""

    def setUp(self):
        """Set up test fixtures."""
        self.config = JiraDownloadConfig(
            company_slug='test',
            url="https://test.atlassian.net",
            gdpr_active=False,
            feature_flags={},
        )

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    def test_legacy_api_selection(self, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test that legacy API is used when JQL Enhanced Search is not available."""
        # Mock API detection to return False (legacy API)
        mock_is_available.return_value = False
        
        # Mock Jira connection
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter
        adapter = JiraCloudManifestAdapter(config=self.config)
        
        # Verify API selection
        self.assertFalse(adapter.use_jql_enhanced_search)
        
        # Verify connection was created with correct parameters
        mock_get_connection.assert_called_once_with(
            config=self.config, use_jql_enhanced_search=False
        )

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    def test_jql_enhanced_search_api_selection(self, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test that JQL Enhanced Search API is used when available and enabled."""
        # Enable JQL Enhanced Search via feature flag
        self.config.feature_flags = {
            Constants.JQL_ENHANCED_SEARCH_ENABLED: True
        }
        
        # Mock API detection to return True (JQL Enhanced Search available)
        mock_is_available.return_value = True
        
        # Mock Jira connection
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter
        adapter = JiraCloudManifestAdapter(config=self.config)
        
        # Verify API selection
        self.assertTrue(adapter.use_jql_enhanced_search)
        
        # Verify connection was created with correct parameters
        mock_get_connection.assert_called_once_with(
            config=self.config, use_jql_enhanced_search=True
        )

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud._get_issue_count_with_jql_enhanced_search')
    def test_get_issues_count_helper_with_jql_enhanced_search(self, mock_enhanced_count, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test _get_issues_count_helper uses JQL Enhanced Search API when enabled."""
        # Enable JQL Enhanced Search
        mock_is_available.return_value = True
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        mock_enhanced_count.return_value = 42
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter with JQL Enhanced Search enabled
        adapter = JiraCloudManifestAdapter(config=self.config)
        adapter.use_jql_enhanced_search = True
        
        # Test _get_issues_count_helper
        result = adapter._get_issues_count_helper("project = TEST")
        
        # Verify correct API was called
        mock_enhanced_count.assert_called_once_with(mock_connection, "project = TEST")
        self.assertEqual(result, 42)

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.retry_for_status')
    def test_get_issues_count_helper_with_legacy_api(self, mock_retry, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test _get_issues_count_helper uses legacy API when JQL Enhanced Search is disabled."""
        # Disable JQL Enhanced Search
        mock_is_available.return_value = False
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        mock_retry.return_value = {'total': 24}
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter with legacy API
        adapter = JiraCloudManifestAdapter(config=self.config)
        adapter.use_jql_enhanced_search = False
        
        # Test _get_issues_count_helper
        result = adapter._get_issues_count_helper("project = TEST")
        
        # Verify correct API was called
        mock_retry.assert_called_once_with(
            mock_connection._get_json,
            'search',
            {'jql': "project = TEST", 'maxResults': 0}
        )
        self.assertEqual(result, 24)

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud._post_raw_result_jql_enhanced')
    def test_get_jql_search_with_jql_enhanced_search(self, mock_enhanced_search, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test _get_jql_search uses JQL Enhanced Search API when enabled."""
        # Enable JQL Enhanced Search
        mock_is_available.return_value = True
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        mock_enhanced_search.return_value = {'issues': [{'id': '123'}]}
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter with JQL Enhanced Search enabled
        adapter = JiraCloudManifestAdapter(config=self.config)
        adapter.use_jql_enhanced_search = True
        
        # Test _get_jql_search
        result = adapter._get_jql_search("project = TEST", max_results=1)
        
        # Verify correct API was called
        mock_enhanced_search.assert_called_once_with(
            jira_connection=mock_connection,
            jql_query="project = TEST",
            fields=['*all'],
            expand=[],
            max_results=1,
            next_page_token=None,
        )
        self.assertEqual(result, {'issues': [{'id': '123'}]})

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.retry_for_status')
    def test_get_jql_search_with_legacy_api(self, mock_retry, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test _get_jql_search uses legacy API when JQL Enhanced Search is disabled."""
        # Disable JQL Enhanced Search
        mock_is_available.return_value = False
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        mock_retry.return_value = {'issues': [{'id': '456'}], 'total': 1}
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter with legacy API
        adapter = JiraCloudManifestAdapter(config=self.config)
        adapter.use_jql_enhanced_search = False
        
        # Test _get_jql_search
        result = adapter._get_jql_search("project = TEST", max_results=1)
        
        # Verify correct API was called
        mock_retry.assert_called_once_with(
            mock_connection._get_json,
            'search',
            {'jql': "project = TEST", 'maxResults': 1}
        )
        self.assertEqual(result, {'issues': [{'id': '456'}], 'total': 1})

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    def test_get_issues_count_uses_jql_count_method(self, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test that get_issues_count uses the new _get_issues_count_helper method."""
        # Mock setup
        mock_is_available.return_value = False
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter
        adapter = JiraCloudManifestAdapter(config=self.config)
        
        # Mock the _get_issues_count_helper method
        adapter._get_issues_count_helper = MagicMock(return_value=100)
        
        # Test get_issues_count
        result = adapter.get_issues_count()
        
        # Verify _get_issues_count_helper was called with correct query
        adapter._get_issues_count_helper.assert_called_once()
        call_args = adapter._get_issues_count_helper.call_args[1]
        self.assertIn('updatedDate >', call_args['jql_query'])
        self.assertEqual(result, 100)

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    def test_get_issues_count_for_project_uses_jql_count_method(self, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test that get_issues_count_for_project uses the new _get_issues_count_helper method."""
        # Mock setup
        mock_is_available.return_value = False
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter
        adapter = JiraCloudManifestAdapter(config=self.config)
        
        # Mock the _get_issues_count_helper method
        adapter._get_issues_count_helper = MagicMock(return_value=50)
        
        # Test get_issues_count_for_project
        result = adapter.get_issues_count_for_project(project_id=123)
        
        # Verify _get_issues_count_helper was called with correct query
        adapter._get_issues_count_helper.assert_called_once()
        call_args = adapter._get_issues_count_helper.call_args[1]
        self.assertIn('project = 123', call_args['jql_query'])
        self.assertIn('updatedDate >', call_args['jql_query'])
        self.assertEqual(result, 50)

    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_projects')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.JiraCloudManifestAdapter._get_all_boards')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.get_jira_connection')
    @patch('jf_ingest.data_manifests.jira.adapters.jira_cloud.is_jql_enhanced_search_available')
    def test_get_last_updated_for_project_still_uses_jql_search(self, mock_is_available, mock_get_connection, mock_get_boards, mock_get_projects):
        """Test that get_last_updated_for_project still uses _get_jql_search for actual issue data."""
        # Mock setup
        mock_is_available.return_value = False
        mock_connection = MagicMock()
        mock_get_connection.return_value = mock_connection
        
        # Mock initialization methods
        mock_get_boards.return_value = []
        mock_get_projects.return_value = []
        
        # Create adapter
        adapter = JiraCloudManifestAdapter(config=self.config)
        
        # Mock the _get_jql_search method to return issue data
        adapter._get_jql_search = MagicMock(return_value={
            'issues': [{
                'fields': {
                    'updated': '2023-05-15T14:04:19.376-0400'
                }
            }]
        })
        
        # Test get_last_updated_for_project
        result = adapter.get_last_updated_for_project(project_id=123)
        
        # Verify _get_jql_search was called with correct parameters
        adapter._get_jql_search.assert_called_once_with(
            jql_search="project = 123 ORDER BY updated",
            max_results=1
        )
        # Verify result is a datetime object
        self.assertIsNotNone(result)


if __name__ == '__main__':
    unittest.main()


from datetime import datetime
from itertools import chain
from typing import Generator

import pytest
import requests
import requests_mock

from jf_ingest.config import AzureDevopsAuthConfig
from jf_ingest.jf_git.clients.azure_devops import AzureDevopsClient

TEST_BASE_URL = 'https://ado.com'
CONTINUATION_TOKEN_HEADER_SLUG = 'x-ms-continuationtoken'


def _get_client(api_version: str = None):
    session = requests.Session()
    auth_config = AzureDevopsAuthConfig(
        company_slug='company_slug',
        token='let-me-in',
        base_url=TEST_BASE_URL,
        session=session,
    )
    if api_version:
        auth_config.api_version = api_version
    return AzureDevopsClient(auth_config)

def test_format_datetime_to_ado_str():
    dt = datetime(2024, 11, 12, 4, 30, 0)
    assert AzureDevopsClient.format_datetime_to_ado_str(dt) == '2024-11-12'

def test_get_json(requests_mock: requests_mock.Mocker):
    def _test_get_json(status_code: int, api_version: str = '7.0', ignore_404: bool = False):
        requests_mock.reset()
        client = _get_client(api_version=api_version)
        # Mock classes/data
        session_mock_return_data = {'data': {'results': [1, 2, 3]}}
        params = {}

        url = f'{TEST_BASE_URL}?api-version={api_version}'
        continuation_token = 'onwards!!!'
        requests_mock.get(
            url,
            headers={CONTINUATION_TOKEN_HEADER_SLUG: continuation_token},
            json=session_mock_return_data,
            status_code=status_code,
        )
        response, returned_continuation_token = client.get_json(
            url=f'{TEST_BASE_URL}', params={}, ignore404=ignore_404
        )
        assert api_version == requests_mock.last_request.qs['api-version'][0]
        if status_code == 404 and ignore_404:
            assert response == {}
            assert returned_continuation_token == None
            return
        assert returned_continuation_token == continuation_token
        assert response == session_mock_return_data

    # Basic smoke test
    _test_get_json(200)
    # Test that API version works with non default value
    _test_get_json(200, api_version='6.4')

    # Test that API version works with non default value
    _test_get_json(404, ignore_404=True)

    with pytest.raises(requests.exceptions.HTTPError):
        _test_get_json(404, ignore_404=False)
        _test_get_json(500)


# NOTE: This test is effectively the same as above
def test_get_single_object(requests_mock: requests_mock.Mocker):
    def _test_get_json(status_code: int, api_version: str = '7.0', ignore_404: bool = False):
        requests_mock.reset()
        client = _get_client(api_version=api_version)
        # Mock classes/data
        session_mock_return_data = {'data': {'results': [1, 2, 3]}}
        params = {}

        url = f'{TEST_BASE_URL}?api-version={api_version}'
        continuation_token = 'onwards!!!'
        requests_mock.get(
            url,
            headers={CONTINUATION_TOKEN_HEADER_SLUG: continuation_token},
            json=session_mock_return_data,
            status_code=status_code,
        )
        response = client.get_single_object(url=f'{TEST_BASE_URL}', params={}, ignore404=ignore_404)
        assert api_version == requests_mock.last_request.qs['api-version'][0]
        if status_code == 404 and ignore_404:
            assert response == {}
            return
        assert response == session_mock_return_data

    # Basic smoke test
    _test_get_json(200)
    # Test that API version works with non default value
    _test_get_json(200, api_version='6.4')

    # Test that API version works with non default value
    _test_get_json(404, ignore_404=True)

    with pytest.raises(requests.exceptions.HTTPError):
        _test_get_json(404, ignore_404=False)
        _test_get_json(500)


def test_get_single_page(requests_mock: requests_mock.Mocker):
    client = _get_client()
    result_container_str = f'date_is_here'
    result_page = ["page", "list", "results"]
    # Mock classes/data
    session_mock_return_data = {result_container_str: result_page}

    url = f'{TEST_BASE_URL}?api-version={client.api_version}'
    requests_mock.get(url, headers={}, json=session_mock_return_data, status_code=200)
    results_from_client = client.get_single_page(url=url, result_container=result_container_str)

    assert result_page == results_from_client


def test_get_single_page_on_ignore_404(requests_mock: requests_mock.Mocker):
    client = _get_client()
    result_container_str = f'date_is_here'
    result_page = ["page", "list", "results"]
    # Mock classes/data
    session_mock_return_data = {result_container_str: result_page}

    url = f'{TEST_BASE_URL}?api-version={client.api_version}'
    requests_mock.get(url, headers={}, json=session_mock_return_data, status_code=404)
    results_from_client = client.get_single_page(
        url=url, result_container=result_container_str, ignore404=True
    )

    assert [] == results_from_client


def test_get_all_pages_using_skip_and_top_one_page(requests_mock: requests_mock.Mocker):
    client = _get_client()

    url = f'{TEST_BASE_URL}'

    result_container = 'value'
    page_results = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    response = {result_container: page_results, 'count': len(page_results)}
    page_size = 100
    requests_mock.get(
        f'{url}?api-version=7.0&%24skip=0&%24top={page_size}',
        headers={},
        json=response,
        status_code=200,
    )

    results = list(
        client.get_all_pages_using_skip_and_top(
            url, result_container=result_container, top=page_size
        )
    )

    assert results == list(page_results)


def test_get_all_pages_using_skip_and_top_multi_page(requests_mock: requests_mock.Mocker):
    client = _get_client()

    url = f'{TEST_BASE_URL}'

    result_container = 'value'
    page_size = 3
    page_one_result = [1, 2, 3]
    page_one_response = {result_container: page_one_result, 'count': len(page_one_result)}
    requests_mock.get(
        f'{url}?api-version=7.0&%24skip=0&%24top={page_size}',
        headers={},
        json=page_one_response,
        status_code=200,
    )
    page_two_results = [4, 5, 6]
    page_two_response = {result_container: page_two_results, 'count': len(page_two_results)}
    requests_mock.get(
        f'{url}?api-version=7.0&%24skip=3&%24top={page_size}',
        headers={},
        json=page_two_response,
        status_code=200,
    )
    page_three_results = [7, 8]
    page_three_response = {result_container: page_three_results, 'count': len(page_three_results)}
    requests_mock.get(
        f'{url}?api-version=7.0&%24skip=6&%24top={page_size}',
        headers={},
        json=page_three_response,
        status_code=200,
    )
    results = list(
        client.get_all_pages_using_skip_and_top(
            url, result_container=result_container, top=page_size
        )
    )

    assert results == list(chain(page_one_result, page_two_results, page_three_results))


def test_get_all_pages_using_skip_and_top_check_assertion_error(
    requests_mock: requests_mock.Mocker,
):
    client = _get_client()

    url = f'{TEST_BASE_URL}'

    result_container = 'value'
    page_results = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    response = {result_container: page_results, 'count': len(page_results)}
    page_size = 100
    requests_mock.get(
        f'{url}?api-version=7.0&%24skip=0&%24top={page_size}',
        headers={CONTINUATION_TOKEN_HEADER_SLUG: 'continuation_token'},
        json=response,
        status_code=200,
    )

    with pytest.raises(AssertionError):
        list(
            client.get_all_pages_using_skip_and_top(
                url, result_container=result_container, top=page_size
            )
        )


def test_get_all_pages_using_pagination_token(requests_mock: requests_mock.Mocker):
    client = _get_client()
    url = f'{TEST_BASE_URL}'
    result_page_one = [1, 2, 3]
    result_page_two = [4, 5, 6]
    result_page_three = [7, 8, 9]

    result_container = 'value'

    def _create_response(result_list):
        return {result_container: result_list}

    requests_mock.get(
        f'{url}?api-version=7.0',
        headers={CONTINUATION_TOKEN_HEADER_SLUG: '1'},
        json=_create_response(result_page_one),
        status_code=200,
    )
    requests_mock.get(
        f'{url}?api-version=7.0&continuationToken=1',
        headers={CONTINUATION_TOKEN_HEADER_SLUG: '2'},
        json=_create_response(result_page_two),
        status_code=200,
    )
    requests_mock.get(
        f'{url}?api-version=7.0&continuationToken=2',
        headers={},
        json=_create_response(result_page_three),
        status_code=200,
    )

    results = list(client.get_all_pages_using_pagination_token(url=url))

    assert results == list(chain(result_page_one, result_page_two, result_page_three))


def test_get_all_pages_using_pagination_token_one_page(requests_mock: requests_mock.Mocker):
    client = _get_client()
    url = f'{TEST_BASE_URL}'
    results = [1, 2, 3, 4, 5, 6, 7, 8, 9]

    result_container = 'value'

    def _create_response(result_list):
        return {result_container: result_list}

    requests_mock.get(
        f'{url}?api-version=7.0', headers={}, json=_create_response(results), status_code=200
    )

    response_results = list(client.get_all_pages_using_pagination_token(url=url))

    assert response_results == results

def test_server_version_not_7_0(requests_mock: requests_mock.Mocker):
    url = f'{TEST_BASE_URL}'
    org_name = 'my_org'

    result_container = 'value'

    def _create_response(result_list):
        return {result_container: result_list}

    def _test_version_number(version_number: str):
        client = _get_client(api_version=version_number)
        requests_mock.get(
            f'{url}/{org_name}/_apis/graph/users',
            headers={},
            json=_create_response([]),
            status_code=200,
        )

        list(client.get_graph_users(org_name='my_org'))

        assert requests_mock.request_history[-1].qs['api-version'][0] == f'{version_number}-preview.1', f"API version should be {version_number}-preview.1"
    
    
    _test_version_number('6.0')
    _test_version_number('6.7')
    _test_version_number('7.0')
    _test_version_number('7.1')
    _test_version_number('8.0')


import logging
from typing import Union

import requests

from jf_ingest import logging_helper
from jf_ingest.config import GitConfig, IngestionConfig
from jf_ingest.jf_git.standardized_models import StandardizedJFAPIPullRequest
from jf_ingest.jf_jira.custom_fields import JELLYFISH_API_TIMEOUT
from jf_ingest.utils import retry_for_status

JF_GHC_PR_USER_ENDPOINT = '/endpoints/agent/github-prs-with-null-users'
JF_GHC_PR_REVIEW_USER_ENDPOINT = '/endpoints/agent/github-pr-reviews-with-null-users'

logger = logging.getLogger(__name__)


def get_jf_github_null_user_pr_data(
    ingest_config: IngestionConfig, git_config: GitConfig, pr_reviews: bool = False
) -> Union[list[StandardizedJFAPIPullRequest], None]:
    """
    Get any PRs or PR reviews with missing mannequin user data in the JF database for a given GHC instance
    """
    headers = {"Jellyfish-API-Token": ingest_config.jellyfish_api_token}
    api_endpoint = JF_GHC_PR_REVIEW_USER_ENDPOINT if pr_reviews else JF_GHC_PR_USER_ENDPOINT

    logger.info(f"Making a request to the Jellyfish API at {api_endpoint}")

    try:
        r = retry_for_status(
            _make_ghc_pr_user_request,
            f"{ingest_config.jellyfish_api_base}{api_endpoint}",
            git_config.instance_slug,
            headers,
            max_retries_for_retry_for_status=2,
            statuses_to_retry=[504],
        )
    except Exception as e:
        # This isn't essential for ingest to continue, so just log the error and return
        logging_helper.send_to_agent_log_file(
            f'Error when attempting to pull PRs with missing mannequin user data from Jellyfish: {e}',
            exc_info=True,
            level=logging.ERROR,
        )
        return None

    try:
        r_json = r.json()
        return [StandardizedJFAPIPullRequest(**pr) for pr in r_json['pull_requests']]
    except Exception as e:
        logging_helper.send_to_agent_log_file(
            f'Error when attempting to parse PRs with missing mannequin user data from Jellyfish: {e}',
            exc_info=True,
            level=logging.ERROR,
        )

    return []


def _make_ghc_pr_user_request(
    api_url: str, git_instance_slug: str, headers: dict[str, str]
) -> requests.Response:
    try:
        r = requests.get(
            f"{api_url}?git-instance-slug={git_instance_slug}",
            headers=headers,
            timeout=JELLYFISH_API_TIMEOUT,
        )
        r.raise_for_status()
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 400:
            logging_helper.send_to_agent_log_file(
                f'GHC instance {git_instance_slug} is not a Github Cloud instance',
                level=logging.ERROR,
            )
            raise e
    except Exception as e:
        logging_helper.send_to_agent_log_file(
            f'Error when attempting to pull data from Jellyfish for endpoint {api_url}',
            exc_info=True,
            level=logging.ERROR,
        )
        raise e

    return r



import ast
import dataclasses
import functools
import json
import logging
import queue
import threading
import time
from collections import defaultdict
from math import ceil
from typing import Any, NamedTuple, Optional, Union

import requests
from opentelemetry import baggage, context, trace

from jf_ingest import logging_helper
from jf_ingest.adaptive_throttler import AdaptiveThrottler
from jf_ingest.config import IngestionConfig
from jf_ingest.jf_jira import (
    get_jira_connection,
    get_jira_search_batch_size,
    pull_jira_issues_by_jira_ids,
)
from jf_ingest.jf_jira.utils import JiraFieldIdentifier, get_jellyfish_jira_issues_count
from jf_ingest.telemetry.tracing import jelly_trace, record_span
from jf_ingest.utils import (
    ThreadPoolExecutorWithLogging,
    init_jf_ingest_run,
    retry_for_status,
    tqdm_to_logger,
)

JELLYFISH_CUSTOM_FIELDS_API_LIMIT = 1_000
JELLYFISH_API_TIMEOUT = 600.0
JELLYFISH_CUSTOM_FIELDS_ENDPOINT = 'endpoints/jira/issues/custom-fields'
JELLYFISH_MARK_FOR_REDOWNLOAD_ENDPOINT = 'endpoints/ingest/jira/issues/mark-for-redownload'
MAX_ISSUES_TO_MARK_FOR_REDOWNLOAD = 400_000

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class JCFVDBData:
    field_id: int
    field_key: str
    field_type: str
    value: str | dict | list


@dataclasses.dataclass
class IssueJCFVDBData:
    issue_jira_id: int
    custom_field_values: list[JCFVDBData]


@dataclasses.dataclass
class JCFVUpdate:
    jira_issue_id: str
    field_id: str
    field_key: str
    field_type: str
    value_old: Optional[dict]
    value_new: Optional[dict]


@functools.lru_cache(maxsize=4096)
def _literal_eval_memoized(value: str) -> dict | list:
    """ast.literal_eval is rather slow, and we expect a lot of duplicate values; memoize it."""
    le: dict = ast.literal_eval(value)
    return le


@functools.lru_cache(maxsize=4096)
def _json_loads_memoized(value: str) -> dict | list:
    """improve performance of JSON loads by memoizing the results."""
    d: dict = json.loads(value)
    return d


def attempt_load_json_or_literal(value: str) -> list[Any] | dict[str, Any] | str:
    """
    Attempt to load a string as JSON or a literal. Used to handle the case where we
    get a dictionary string back that's out of order.

    Args:
        value (str): The string to attempt to load.

    Raises:
    """
    if isinstance(value, str):
        if value.startswith('{') or value.startswith('['):
            try:
                return _json_loads_memoized(value)
            except Exception:
                try:
                    return _literal_eval_memoized(value)
                except Exception:
                    pass
    return value


@jelly_trace
def _annotate_results_from_jellyfish(data: dict) -> dict[str, IssueJCFVDBData]:
    """
    Unpacks the results from Jellyfish into a dict[str, IssueJCFVDBData] - where `str` is the jira_issue_id.
    Typically, the value being sent over is JSON, so we need to handle that here.
    """
    resp: dict[str, IssueJCFVDBData] = {}
    for issue in data['issues']:
        resp[str(issue['issue_jira_id'])] = IssueJCFVDBData(
            issue_jira_id=issue['issue_jira_id'], custom_field_values=[]
        )
        for jcfv_db_data in issue['custom_field_values']:
            value = attempt_load_json_or_literal(jcfv_db_data['value'])
            resp[str(issue['issue_jira_id'])].custom_field_values.append(
                JCFVDBData(
                    field_id=jcfv_db_data['field_id'],
                    field_key=jcfv_db_data['field_key'],
                    field_type=jcfv_db_data['field_type'],
                    value=value,
                )
            )
    return resp


def _fetch_custom_fields_from_jellyfish(
    base_url: str, headers: dict[str, str], cursor: int, limit: int
) -> requests.Response:
    try:
        response = requests.get(
            f"{base_url}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?cursor={cursor}&limit={limit}",
            headers=headers,
            timeout=JELLYFISH_API_TIMEOUT,
        )
        response.raise_for_status()
    except Exception as e:
        logging_helper.send_to_agent_log_file(
            f'Error when attempting to pull custom fields from Jellyfish',
            exc_info=True,
            level=logging.ERROR,
        )
        raise e

    return response


@jelly_trace
def _retrieve_custom_fields_from_jellyfish(
    ingest_config: IngestionConfig,
    output_queue: queue.Queue,
    max_issues_to_process: Optional[int] = None,
    limit: int = JELLYFISH_CUSTOM_FIELDS_API_LIMIT,
    exit_event: threading.Event = threading.Event(),
) -> None:
    """
    Retrieve all custom fields from Jellyfish.
    Note: 1K is the API hard limit.
    """
    if max_issues_to_process == 0:
        return

    cursor = 0
    total_fetched_issues = 0
    limit = min((limit, max_issues_to_process)) if max_issues_to_process else limit

    base_url = ingest_config.jellyfish_api_base
    headers = {"Jellyfish-API-Token": ingest_config.jellyfish_api_token}

    while True:
        if exit_event.is_set():
            while not output_queue.empty():
                try:
                    output_queue.get_nowait()
                except queue.Empty:
                    break

            return None

        st = time.perf_counter()

        # Sometimes we'll see 504 from Jellyfish (e.g. if pulling a large amount of data) so we'll retry for status here
        r = retry_for_status(_fetch_custom_fields_from_jellyfish, base_url, headers, cursor, limit)
        data = r.json()

        # Catch if the API changes the maximum allowed limit compared to what was provided
        limit = min((limit, data['max_records']))

        # If we get 0 for cursor, there is no more data to fetch, otherwise we fetched the limit
        cursor = int(data['next_cursor'])

        if not cursor:
            fetched_count = len(data['issues'])
        else:
            fetched_count = limit
        total_fetched_issues += fetched_count

        # We should only queue responses that have data.
        if fetched_count:
            # Annotate data to be put into the output queue.
            output_queue.put(_annotate_results_from_jellyfish(data))

        if not data['issues']:
            logging_helper.send_to_agent_log_file(
                f'No more issues found when attempting to pull custom fields from Jellyfish. {total_fetched_issues} issues retrieved.',
            )
            break

        if max_issues_to_process is not None and total_fetched_issues >= max_issues_to_process:
            logging_helper.send_to_agent_log_file(
                f'Finished pulling custom fields from Jellyfish - reached max_issues_to_process. {total_fetched_issues} issues retrieved.'
            )
            break

        if not cursor:
            logging_helper.send_to_agent_log_file(
                f'Finished pulling custom fields from Jellyfish - reached end of data. {total_fetched_issues} issues retrieved.'
            )
            break

        logging_helper.send_to_agent_log_file(
            f'Pulled {total_fetched_issues} issues from Jellyfish, {fetched_count} records in {time.perf_counter() - st:.2f}s,'
            f' continuing...',
            level=logging.DEBUG,
        )


class JCFVUpdateFullPayload(NamedTuple):
    missing_from_db_jcfv: list[JCFVUpdate]
    missing_from_jira_jcfv: list[JCFVUpdate]
    out_of_sync_jcfv: list[JCFVUpdate]


def _send_marked_for_redownload_to_jellyfish(base_url, headers, data) -> requests.Response:
    try:
        response = requests.post(
            f"{base_url}/{JELLYFISH_MARK_FOR_REDOWNLOAD_ENDPOINT}",
            headers=headers,
            json=data,
            timeout=JELLYFISH_API_TIMEOUT,
        )
        response.raise_for_status()
    except Exception as e:
        logging_helper.send_to_agent_log_file(
            f'Error when attempting to mark Jira issues for redownload via Jellyfish',
            exc_info=True,
            level=logging.ERROR,
        )
        raise e

    return response


def make_hash(o: Union[set, list, tuple, dict, Any]) -> Union[int, tuple, None]:
    """
    Makes a hash from a dictionary, list, tuple or set to any level, that contains
    only other hashable types (including any lists, tuples, sets, and
    dictionaries).
    """

    if isinstance(o, (set, tuple, list)):
        return tuple(sorted([make_hash(e) for e in o]))

    elif not isinstance(o, dict):
        return hash(o)
    o_new = dict()
    for k, v in o.items():
        o_new[k] = make_hash(v)

    return hash(tuple(frozenset(sorted(o_new.items()))))


@jelly_trace
def identify_custom_field_mismatches(
    ingest_config: IngestionConfig,
    use_jql_enhanced_search: bool = False,
    nthreads: int = 10,
    max_issues_to_process: Optional[int] = None,
    mark_for_redownload: Optional[int] = None,
    use_throttler: bool = False,
) -> JCFVUpdateFullPayload:
    """
    At a high level:
    - Get all of the fields that are used to define critical components in Jellyfish. We use a Jellyfish
        endpoint to get this data, in order to support using this sync path from both the agent and internally.
    - Create one thread that retrieves (issue.id, issue.jira_id, *custom_field_values) of all issues
        and puts that into a queue. This is done in a separate thread to allow us to download issues from Jira
        in parallel, as downloading issues from Jira is the primary bottleneck.
    - In our main thread, read from the queue that's being stuffed by the _retrieve_custom_fields_from_jellyfish
        thread, and query Jira to get up-to-date issue data. If there is any data mismatch in custom
        fields, save that into an "updates" NamedTuple, which we can POST to Jellyfish to update the data.

    :param ingest_config: IngestionConfig object
    :param use_jql_enhanced_search: Whether to use JQL Enhanced Search API (/search/jql) instead of legacy API (/search)
    :param nthreads: Number of threads to use for downloading issues from Jira
    :param max_issues_to_process: Maximum number of issues to process. If None, all issues will be processed.
    :param mark_for_redownload: Send jira issue ids to mark them for redownload in Jellyfish.
    :param use_throttler: Whether to use the adaptive throttler for Jira API requests.
    :return: Tuple of (missing_from_db_jcfv, missing_from_jira_jcfv, out_of_sync_jcfv)
    """
    init_jf_ingest_run(ingestion_config=ingest_config)
    logger.info(f'Attempting to identify custom field mismatches between Jira and Jellyfish')
    total_issue_count = get_jellyfish_jira_issues_count(
        jellyfish_api_base_url=ingest_config.jellyfish_api_base,
        jellyfish_api_token=ingest_config.jellyfish_api_token,
    )
    output_queue: queue.Queue = queue.Queue()
    exit_event = threading.Event()

    # Use ThreadPoolExecutor here so that we can capture
    # any exceptions raised within the thread.
    executor = ThreadPoolExecutorWithLogging(max_workers=1)
    retrieval_thread = executor.submit(
        _retrieve_custom_fields_from_jellyfish,
        ingest_config,
        output_queue,
        max_issues_to_process=max_issues_to_process,
        limit=JELLYFISH_CUSTOM_FIELDS_API_LIMIT,
        exit_event=exit_event,
    )

    # We _are_ downloading issues but not doing a full download, so for_download=False
    # is fine here and is faster.
    if not (jira_config := ingest_config.jira_config):
        raise Exception(f'Ingest config was provided without a Jira Configuration!')
    jira_connection = get_jira_connection(
        config=jira_config, use_jql_enhanced_search=use_jql_enhanced_search
    )

    # We store the custom field values that are missing from the DB, missing from Jira,
    # and out of sync in separate lists - this allows for easier processing on the
    # update side.
    update_payload = JCFVUpdateFullPayload([], [], [])

    # Use the adaptive throttler for Jira server/dc requests if enabled
    jira_request_throttler = (
        AdaptiveThrottler(logging_extra={'company_slug': ingest_config.company_slug})
        if use_throttler
        else None
    )

    st = time.perf_counter()
    total_issues_scanned = 0
    progress_bar = tqdm_to_logger(
        desc=f'Detecting changes to custom fields in Jira...', total=total_issue_count
    )
    while retrieval_thread.running() or not output_queue.empty():
        if exit_event.is_set():
            break

        try:
            db_issue_batch: dict[str, IssueJCFVDBData] = output_queue.get(timeout=60.0)
        except queue.Empty:
            logging_helper.send_to_agent_log_file(
                'Didn\'t get any issues from the queue in 60 seconds, retrying...'
            )
            continue

        total_issues_scanned += len(db_issue_batch)

        # Get all the field keys from the DB data, so we can request them from Jira.
        field_map: dict[str, tuple[int, str, str]] = {}
        for issue in db_issue_batch.values():
            for jcfv in issue.custom_field_values:
                field_map[jcfv.field_key] = (jcfv.field_id, jcfv.field_key, jcfv.field_type)

        # If we don't have any fields to query, we can skip the Jira query.
        if not field_map:
            continue

        # Likely won't change the batch size vs. doing this without "fields", but
        # it's possible we'll be able to use a larger batch size if querying a subset
        # of fields.
        batch_size = get_jira_search_batch_size(
            jira_connection,
            fields=['key'] + list(field_map.keys()),
            use_jql_enhanced_search=use_jql_enhanced_search,
        )

        # Note: this is parallelized and returns a generator, so we're
        # able to process issues while downloading from Jira simultaneously.
        jira_ids_to_search = list(db_issue_batch.keys())
        include_fields = [
            JiraFieldIdentifier(jira_field_id=key, jira_field_name='')
            for key in ['key'] + list(field_map.keys())
        ]
        with record_span('jira_issues_download'):
            jira_issue_batch = pull_jira_issues_by_jira_ids(
                jira_connection=jira_connection,
                jira_ids=jira_ids_to_search,
                num_parallel_threads=nthreads,
                batch_size=batch_size,
                include_fields=include_fields,
                hide_tqdm=True,
                adaptive_throttler=jira_request_throttler,
                use_jql_enhanced_search=use_jql_enhanced_search,
            )

        # Now we need to compare the custom field values from the DB with the actual issue data from Jira.
        tracer = trace.get_tracer(__name__)
        compare_issues_to_db_span = tracer.start_span(name='compare_issues_to_db')
        for issue_jira in jira_issue_batch:
            issue_db = db_issue_batch[str(issue_jira['id'])]
            jcfv_db_dict = {jcfv.field_key: jcfv for jcfv in issue_db.custom_field_values}

            for field_id, field_key, field_type in field_map.values():
                # Attempt to get the custom field value from the DB.
                # This may not be present, in the case that the field was
                # added to the issue, but we haven't yet pulled the data.
                db_value = getattr(jcfv_db_dict.get(field_key), 'value', None)

                # Occasionally, an issue may not have any of the fields we
                # requested, (e.g. if we're looking for just one field which is
                # no longer present on the issue present) so we need to handle this case.
                jira_value = issue_jira.get('fields', {}).get(field_key, None)
                hash_db_value = make_hash(db_value)
                hash_jira_value = make_hash(jira_value)
                if hash_db_value != hash_jira_value:
                    if db_value is None:
                        list_to_append = update_payload.missing_from_db_jcfv
                    elif jira_value is None:
                        list_to_append = update_payload.missing_from_jira_jcfv
                    else:
                        list_to_append = update_payload.out_of_sync_jcfv
                    list_to_append.append(
                        JCFVUpdate(
                            field_id=str(field_id),
                            field_key=field_key,
                            field_type=field_type,
                            jira_issue_id=str(issue_db.issue_jira_id),
                            value_old=db_value,
                            value_new=jira_value,
                        )
                    )
        compare_issues_to_db_span.end()

        total_fields_out_of_sync = sum(map(len, update_payload))
        logging_helper.send_to_agent_log_file(
            f'Downloaded {total_issues_scanned} issues from Jira. '
            f'{total_fields_out_of_sync} out of sync field values found '
            f'so far in {time.perf_counter() - st} seconds.'
        )
        st = time.perf_counter()
        progress_bar.update(len(jira_ids_to_search))

        if total_fields_out_of_sync >= MAX_ISSUES_TO_MARK_FOR_REDOWNLOAD:
            logging_helper.send_to_agent_log_file(
                f'Exceeded the maximum number of issues to mark for redownload. Exiting early. '
                f'{total_fields_out_of_sync} out of sync field values found in total. '
                f'Max issues to mark for redownload: {MAX_ISSUES_TO_MARK_FOR_REDOWNLOAD}'
            )
            exit_event.set()

    # Run here in case we get no issues back from Jellyfish
    total_fields_out_of_sync = sum(map(len, update_payload))
    logging_helper.send_to_agent_log_file(
        f'Finished processing all issues from the DB. {total_fields_out_of_sync} '
        f'out of sync field values found in total.'
    )
    logging_helper.send_to_agent_log_file(
        f'{len(update_payload.out_of_sync_jcfv)} custom field values to update.'
    )
    logging_helper.send_to_agent_log_file(
        f'{len(update_payload.missing_from_jira_jcfv)} custom field values to delete.'
    )
    logging_helper.send_to_agent_log_file(
        f'{len(update_payload.missing_from_db_jcfv)} custom field values to insert.'
    )
    logging_helper.send_to_agent_log_file(f'{total_issues_scanned} issues scanned in total.')

    # Retrieval thread is already finished, as guaranteed by the while loop above.
    # We can safely call result() here, which will raise any exceptions we might have seen.
    retrieval_thread.result()

    # Logging
    update_counts: defaultdict[Any, int] = defaultdict(int)
    for update in (
        update_payload.out_of_sync_jcfv
        + update_payload.missing_from_db_jcfv
        + update_payload.missing_from_jira_jcfv
    ):
        update_counts[f'{update.value_old} -> {update.value_new}'] += 1
    for value_change, update_count in sorted(
        update_counts.items(), key=lambda x: x[1], reverse=True
    ):
        if value_change.startswith('None -> '):
            prefix = 'INSERT'
        elif value_change.endswith(' -> None'):
            prefix = 'DELETE'
        else:
            prefix = 'UPDATE'
        logging_helper.send_to_agent_log_file(f'{update_count: >6d} | {prefix} | {value_change }')

    if mark_for_redownload:
        base_url = ingest_config.jellyfish_api_base
        headers = {"Jellyfish-API-Token": ingest_config.jellyfish_api_token}

        # These need to be cast to strings, since the endpoint only accepts issue ids as string values
        missing_db_ids = [str(jcfv.jira_issue_id) for jcfv in update_payload.missing_from_db_jcfv]
        missing_jira_ids = [
            str(jcfv.jira_issue_id) for jcfv in update_payload.missing_from_jira_jcfv
        ]
        missing_out_of_sync_ids = [
            str(jcfv.jira_issue_id) for jcfv in update_payload.out_of_sync_jcfv
        ]
        all_ids = set(missing_db_ids + missing_jira_ids + missing_out_of_sync_ids)

        if len(all_ids) >= MAX_ISSUES_TO_MARK_FOR_REDOWNLOAD:
            all_ids = set(list(all_ids)[:MAX_ISSUES_TO_MARK_FOR_REDOWNLOAD])

        if all_ids:
            logging_helper.send_to_agent_log_file(
                f'Submitting {len(all_ids)} issue IDs to Jellyfish that were marked for redownload'
            )

            batch_size = 10_000
            total_batches = ceil(len(all_ids) / batch_size)
            id_list = list(all_ids)

            for i in range(total_batches):
                end_idx = min((i + 1) * batch_size, len(all_ids))
                data = {'issue_ids': id_list[i * batch_size : end_idx]}
                retry_for_status(_send_marked_for_redownload_to_jellyfish, base_url, headers, data)
                logging_helper.send_to_agent_log_file(
                    f'Batch {i} of {total_batches} submitted for redownload'
                )

            logging_helper.send_to_agent_log_file(
                f'Done submitting issues for repair. {len(all_ids)} issues were marked for redownload'
            )
        else:
            logging_helper.send_to_agent_log_file(
                'No Jira Issues were marked as needing their custom fields repaired'
            )

    return update_payload




from dataclasses import asdict
from unittest.mock import patch
import pytest
from jf_ingest.config import GitAuthConfig, GitConfig, GitProvider, GithubAuthConfig
from jf_ingest.jf_git.adapters import GitAdapter
from jf_ingest.jf_git.adapters.github import GithubAdapter
from jf_ingest.jf_git.clients.github import GithubClient
from jf_ingest.jf_git.standardized_models import StandardizedOrganization
from tests.jf_git.github.utils import (
    TEST_COMPANY_SLUG,
    TEST_BASE_URL,
    TEST_BASE_GQL_URL,
    TEST_INSTANCE_FILE_KEY,
    TEST_INSTANCE_SLUG,
    TEST_TOKEN
)

TEST_GIT_AUTH_CONFIG = GitAuthConfig(
        company_slug=TEST_COMPANY_SLUG,
        base_url=TEST_BASE_URL,
        token=TEST_TOKEN,
        verify=False,
    )
BASE_ORGANIZATION = StandardizedOrganization(
    id='org_login',
    name='An Organizaiton',
    login='org_login',
    url='www.organization.com'
)

GITHUB_CONFIG = GitConfig(
    company_slug=TEST_COMPANY_SLUG,
    instance_slug=TEST_INSTANCE_SLUG,
    instance_file_key=TEST_INSTANCE_FILE_KEY,
    git_provider=GitProvider.GITHUB,
    git_auth_config=TEST_GIT_AUTH_CONFIG,
    git_organizations=[BASE_ORGANIZATION.login]
)

BASE_ORGANIZATION_FULL_PATH = 'group_full_path'


def _get_github_adapter() -> GithubAdapter:
    adapter: GithubAdapter = GitAdapter.get_git_adapter(GITHUB_CONFIG)
    return adapter

def test_github_adapter():
    gitlab_adapter = _get_github_adapter()
    assert type(gitlab_adapter) == GithubAdapter
    assert gitlab_adapter.config.git_provider == GitProvider.GITHUB
    assert type(gitlab_adapter.client) == GithubClient
    
TEST_BASE_URL = 'https://gav.com/gql'

# Shared test data
GQL_DATETIME_STR = "2024-12-29T12:00:58Z"
    
@pytest.fixture
def github_adapter():
    print(f'Getting adapter')
    return _get_github_adapter()

def test_get_api_scopes(github_adapter: GithubAdapter):
    ret_val = '[user:email, repo, user, organization]'
    with patch.object(GithubClient, 'get_scopes_of_api_token', return_value=ret_val):
        github_adapter.get_api_scopes() == ret_val
        
def test_get_api_scopes(github_adapter: GithubAdapter):
    ret_val = '[user:email, repo, user, organization]'
    with patch.object(GithubClient, 'get_scopes_of_api_token', return_value=ret_val):
        github_adapter.get_api_scopes() == ret_val
        
def test_get_organizations_discover(github_adapter: GithubAdapter):
    github_adapter.config.discover_organizations = True
    raw_orgs = [
        {
            'id': 'id_1',
            'login': 'login_1',
            'name': 'name_1',
            'url': 'www.url-1.com'
        },
        {
            'id': 'id_2',
            'login': 'login_2',
            'name': 'name_2',
            'url': 'www.url-2.com'
        },
    ]
    with patch.object(GithubClient, 'get_all_organizations', return_value=raw_orgs):
        standardized_orgs = github_adapter.get_organizations()
        assert len(standardized_orgs) == len(raw_orgs)
        for org, raw_org in zip(standardized_orgs, raw_orgs):
            assert type(org) == StandardizedOrganization
            assert org.id == raw_org['id']
            assert org.login == raw_org['login']
            assert org.name == raw_org['name']
            assert org.url == raw_org['url']
        
def test_get_organizations_not_discover(github_adapter: GithubAdapter):
    github_adapter.config.discover_organizations = False
    with patch.object(GithubClient, 'get_organization_by_login', return_value=asdict(BASE_ORGANIZATION)):
        standardized_orgs = github_adapter.get_organizations()
        assert len(standardized_orgs) == 1
        org = standardized_orgs[0]
        assert type(org) == StandardizedOrganization
        assert org.id == BASE_ORGANIZATION.id
        assert org.login == BASE_ORGANIZATION.login
        assert org.name == BASE_ORGANIZATION.name
        assert org.url == BASE_ORGANIZATION.url


import json
import logging
import time
import traceback
import zoneinfo
from dataclasses import dataclass
from datetime import datetime
from types import MappingProxyType
from typing import Optional, Union

from jira import JIRAError

from jf_ingest.config import JiraDownloadConfig
from jf_ingest.constants import Constants
from jf_ingest.data_manifests.jira.adapters.manifest_adapter import ManifestAdapter
from jf_ingest.data_manifests.manifest_base import ManifestSource
from jf_ingest.jf_jira import get_jira_connection
from jf_ingest.jf_jira.downloaders import (
    _get_issue_count_with_jql_enhanced_search,
    _post_raw_result_jql_enhanced,
    is_jql_enhanced_search_available,
    search_users,
)
from jf_ingest.utils import get_wait_time, retry_for_status

logger = logging.getLogger(__name__)


@dataclass
class JiraCloudManifestAdapter(ManifestAdapter):
    '''
    Adapter for retrieving information from all remote Jira sources
    '''

    def __init__(
        self,
        config: JiraDownloadConfig,
        project_keys_to_classification_type: Union[
            dict[str, str], MappingProxyType
        ] = MappingProxyType({}),
    ):
        # Super class fields
        super().__init__(
            config=config,
            manifest_source=ManifestSource.remote,
            project_keys_to_classification_type=project_keys_to_classification_type,
        )
        # Calculate API selection decision using feature flags
        use_jql_enhanced_search = is_jql_enhanced_search_available(
            jira_config=config,
            jql_enhanced_search_enabled=config.feature_flags.get(
                Constants.JQL_ENHANCED_SEARCH_ENABLED, False
            ),
            force_legacy_api=config.feature_flags.get(Constants.FORCE_LEGACY_API, False),
        )

        # Store API selection decision as instance variable
        self.use_jql_enhanced_search = use_jql_enhanced_search

        self.jira_connection = get_jira_connection(
            config=config, use_jql_enhanced_search=use_jql_enhanced_search
        )

        # Rip off trailing slash since all urls depend on that
        self.jira_endpoint = (
            str(self.config.url)[:-1] if str(self.config.url).endswith('/') else self.config.url
        )

        self._boards_cache: list = []
        self._projects_cache: list = []

        # Set up caches
        self._get_all_boards()
        self._get_all_projects()

    def get_users_count(self) -> int:
        # Users are a special case where their page results aren't formatted the same as
        # other API page results. Special, internal function to handle this
        def _get_user_count_from_pages(url_path: str, page_size: int = 50) -> int:
            curr = 0
            user_count = 0
            while True:
                page = retry_for_status(
                    self.jira_connection._get_json,
                    url_path,
                    {**{'startAt': curr, "maxResults": page_size}},
                )
                page_count = len(page)
                user_count += page_count
                curr += page_size

                if page_count == 0:
                    break
            return user_count

        # Jira Cloud
        if self.config.gdpr_active:
            return _get_user_count_from_pages(url_path='users/search')
        # Jira Server
        else:
            return len(
                search_users(
                    jira_connection=self.jira_connection,
                    gdpr_active=bool(self.config.gdpr_active),
                    search_users_by_letter_email_domain=self.config.search_users_by_letter_email_domain,
                )
            )

    def get_fields_count(self) -> int:
        # Lazy loading paranoia, we might not need to do this for loop
        return len(retry_for_status(self.jira_connection.fields))

    def get_resolutions_count(self) -> int:
        # Lazy loading paranoia, we might not need to do this for loops
        try:
            return len(retry_for_status(self.jira_connection.resolutions))
        except Exception as e:
            logger.error(f"Could not get resolutions count for {self.config}, got {e}")
            return 0

    def get_issue_types_count(self) -> int:
        # Lazy loading paranoia, we might not need to do this for loop
        return len(retry_for_status(self.jira_connection.issue_types))

    def get_issue_link_types_count(self) -> int:
        return len(retry_for_status(self.jira_connection.issue_link_types))

    def get_priorities_count(self) -> int:
        return len(retry_for_status(self.jira_connection.priorities))

    def get_boards_count(self) -> int:
        return len(retry_for_status(self._get_all_boards))

    def get_project_versions_count(self) -> int:
        total_project_versions = 0
        retries = 0
        max_retries = 5
        constant_sleep = 5
        for project in self._get_all_projects():
            done = False
            while not done:
                try:
                    total_project_versions += self._get_raw_result(
                        url=f'{self.jira_endpoint}/rest/api/latest/project/{project["id"]}/version?startAt=0&maxResults=0&state=future,active,closed'
                    )['total']
                    retries = 0
                    done = True
                except Exception as e:
                    if (
                        hasattr(e, 'status_code')
                        and e.status_code == 429
                        and retries <= max_retries
                    ):
                        sleep_secs = get_wait_time(e, retries)
                        logger.info(
                            f'Caught JIRAError with a {e.status_code} return code during get_project_versions_count. '
                            f'Hopefully transient; sleeping for {sleep_secs} secs then may retry ({retries} of {max_retries}).'
                        )
                        time.sleep(sleep_secs + (constant_sleep * retries))
                        retries += 1
                    else:
                        logger.error(
                            f"Could not recover from error getting jira projects for manifests, got {e}"
                        )
                        raise

        return total_project_versions

    def _get_all_projects(self) -> list[dict]:
        if not self._projects_cache:
            if self.config.gdpr_active:
                self._projects_cache = [
                    project
                    for project in self._page_get_results(
                        url=f'{self.jira_endpoint}/rest/api/latest/project/search?startAt=%s&maxResults=500&status=live'
                    )
                    if not project.get('archived', False) and not project['isPrivate']
                ]
            else:
                self._projects_cache = [
                    project
                    for project in self._get_raw_result(
                        url=f'{self.jira_endpoint}/rest/api/latest/project?includedArchived=True'
                    )
                ]

        return self._projects_cache

    def get_project_data_dicts(self) -> list[dict]:
        return [
            {"key": project['key'], "id": project['id']} for project in self._get_all_projects()
        ]

    def get_project_versions_count_for_project(self, project_id: int) -> int:
        return (
            self._get_raw_result(
                url=f'{self.jira_endpoint}/rest/api/latest/project/{project_id}/version?startAt=0&maxResults=0&state=future,active,closed'
            )['total']
            or 0
        )

    def get_issues_count(self) -> int:
        # JCR 072123 Since we're using JiraIssueData, we need to check `download_all_json`, since this
        # determines whether we get _all_ issue data or just data from the pull_from date.
        #
        # See: `download_all_issue_metadata`
        pull_from = (
            '0' if self.config.full_redownload else self.config.pull_from.strftime('%Y-%m-%d')
        )

        # Query for all issues via JQL to get count
        # Use dedicated count method for clean separation of concerns
        return self._get_issues_count_helper(jql_query=f"updatedDate > {pull_from}")

    def get_board_ids(self, project_id: int):
        result = self._get_raw_result(
            url=f'{self.jira_endpoint}/rest/agile/1.0/board?projectKeyOrId={project_id}&startAt=0&maxResults=100'
        )
        return [value['id'] for value in result['values']]

    def _get_all_boards(self):
        if not self._boards_cache:
            self._boards_cache = [
                board
                for board in self._page_get_results(
                    url=f'{self.jira_endpoint}/rest/agile/1.0/board?startAt=%s&maxResults=100'
                )
            ]

        return self._boards_cache

    def test_basic_auth_for_project(self, project_id: int) -> bool:
        # Doing a basic query for issues is the best way to test auth.
        # Catch and error, if it happens, and bubble up more specific error
        try:
            self.get_issues_count_for_project(project_id=project_id)
            return True
        except JIRAError:
            return False
        except Exception:
            # This is unexpected behavior and it should never happen, log the error
            # before returning
            logger.error(
                'Unusual exception encountered when testing auth. '
                f'JIRAError was expected but the following error was raised: {traceback.format_exc()}'
            )
            return False

    def get_issues_count_for_project(self, project_id: int) -> int:
        # JCR 072123 Since we're using JiraIssueData, we need to check `download_all_json`, since this
        # determines whether we get _all_ issue data or just data from the pull_from date.
        #
        # See: `download_all_issue_metadata`
        pull_from = (
            '0' if self.config.full_redownload else self.config.pull_from.strftime('%Y-%m-%d')
        )

        # Query for all issues via JQL to get count
        # Use dedicated count method for clean separation of concerns
        return self._get_issues_count_helper(
            jql_query=f"project = {project_id} AND updatedDate > {pull_from}"
        )

    # I believe that there should be a one to one relationship with
    # Jira Issues and Jira Issue Data...
    def get_issues_data_count_for_project(self, project_id: int) -> int:
        return self.get_issues_count_for_project(project_id=project_id)

    def _get_issues_count_helper(self, jql_query: str) -> int:
        """Get issue count using appropriate API based on instance configuration."""
        try:
            if self.use_jql_enhanced_search:
                return _get_issue_count_with_jql_enhanced_search(self.jira_connection, jql_query)
            else:
                # Legacy API: get count from 'total' field
                response = retry_for_status(
                    self.jira_connection._get_json,
                    'search',
                    {'jql': jql_query, 'maxResults': 0},
                )
                return int(response['total'] or 0)
        except Exception as e:
            logger.error(f"Could not recover from error getting jira count for manifests, got {e}")
            raise

    def _get_jql_search(self, jql_search: str, max_results: int = 0):
        """Get search results using appropriate API based on instance configuration."""
        retries = 0
        max_retries = 5
        constant_sleep = 5
        while True:
            try:
                if self.use_jql_enhanced_search:
                    # JQL Enhanced Search API: use _post_raw_result_jql_enhanced
                    return _post_raw_result_jql_enhanced(
                        jira_connection=self.jira_connection,
                        jql_query=jql_search,
                        fields=['*all'],
                        expand=[],
                        max_results=max_results,
                        next_page_token=None,
                    )
                else:
                    # Legacy API: existing behavior
                    return retry_for_status(
                        self.jira_connection._get_json,
                        'search',
                        {'jql': jql_search, "maxResults": max_results},
                    )
            except Exception as e:
                if hasattr(e, 'status_code') and e.status_code == 429 and retries <= max_retries:
                    sleep_secs = get_wait_time(e, retries)
                    logger.info(
                        f'Caught JIRAError with a {e.status_code} return code during jira_connection._get_json. '
                        f'Hopefully transient; sleeping for {sleep_secs} secs then may retry ({retries} of {max_retries}).'
                    )
                    time.sleep(sleep_secs + (constant_sleep * retries))
                    retries += 1
                else:
                    logger.error(
                        f"Could not recover from error getting jira search for manifests, got {e}"
                    )
                    raise

    def _page_get_results(self, url: str):
        start_at = 0
        while True:
            page_result = self._get_raw_result(url % start_at)
            for value in page_result['values']:
                yield value

            if page_result['isLast']:
                break
            else:
                start_at += len(page_result['values'])

    def _get_raw_result(self, url) -> dict:
        response = retry_for_status(self.jira_connection._session.get, url)
        response.raise_for_status()
        json_str = response.content.decode()
        return json.loads(json_str) or {}

    def get_last_updated_for_project(self, project_id: int) -> Optional[datetime]:
        try:
            most_recent_issue_from_api: dict = self._get_jql_search(
                jql_search=f"project = {project_id} ORDER BY updated",
                max_results=1,
            )['issues'][0]
            # Example returned str: '2023-05-15T14:04:19.376-0400'
            most_recent_updated_date_str = most_recent_issue_from_api['fields']['updated']
            last_updated_datetime = datetime.strptime(
                most_recent_updated_date_str, '%Y-%m-%dT%H:%M:%S.%f%z'
            )
            # Normalize to utc
            return last_updated_datetime.astimezone(tz=zoneinfo.ZoneInfo('UTC'))
        except Exception as e:
            logger.error(
                f"Could not fetch latest issue updated date in project {project_id}, got: {e}"
            )
            return None

    """
    NOTE: DEPRECATED FUNCTION
    This function takes a long time to process and we currently do not use sprint counts for anything.
    I have made the decision to deprecate it, but there is no standard decorator for deprecated functions in python.
    Instead, I have removed this function from the parent abstract class and commented it out.
    I have opted to comment it out because it's a non trivial function, and if we want sprint counts back
    sometime in the future, it'd be nice to have this to work off of
    -Gavin
    
    NOTE: If we do bring this back, we must add retry logic with exponential back off logic to get around 429 errors!
    
    def get_sprints_count(self) -> int:
        # Leverage the jira_connections _session field to get an authenticated session
        # Get all boards with above query, get all sprints with this query: https://docs.atlassian.com/jira-software/REST/7.3.1/?#agile/1.0/board/{boardId}/sprin

        def _get_sprints_for_board(board_id: int):
            total_sprints_for_board = 0
            try:
                total_sprints_for_board += len(
                    [
                        sprint
                        for sprint in self._page_get_results(
                            url=f'{self.jira_endpoint}/rest/agile/1.0/board/{board_id}/sprint?startAt=%s&maxResults=100'
                        )
                    ]
                )
            except JIRAError as e:
                # JIRA returns 500 errors for various reasons: board is
                # misconfigured; "failed to execute search"; etc.  Just
                # skip and move on
                if e.status_code == 500:
                    logger.debug(f"Couldn't get sprints for {self.instance} board {board_id}")
                elif e.status_code == 400:
                    logger.debug(
                        f"For {self.instance}, board {board_id} doesn't support sprints -- skipping"
                    )
                else:
                    raise

            return total_sprints_for_board

        with ThreadPoolExecutor(max_workers=5) as executor:
            futures = [
                executor.submit(_get_sprints_for_board, board["id"])
                for board in self._get_all_boards()
            ]

            return sum([future.result() for future in futures if future])
    """



import re
from typing import Optional


class NameRedactor:
    """This is a helper class for cleaning up sensitive data that gets submitted
    by the agent. It is currently only used by the Git ingest functionality.

    The boolean fields related to this logic are found on the GitConfig. They are:
        GitConfig.git_strip_text_content,
        GitConfig.git_redact_names_and_urls,
    """

    def __init__(self, preserve_names: Optional[list[str]] = None) -> None:
        self.redacted_names: dict[str, str] = {}
        self.seq = 0
        self.preserve_names: list[str] = preserve_names or []

    def redact_name(self, name: Optional[str]) -> Optional[str]:
        if not name or name in self.preserve_names:
            return name

        redacted_name = self.redacted_names.get(name)

        # Check to see if name is already redacted
        if re.match('redacted-\d\d\d\d', name):
            redacted_name = name

        if not redacted_name:
            redacted_name = f'redacted-{self.seq:04}'
            self.seq += 1
            self.redacted_names[name] = redacted_name
        return redacted_name
